import re
import ast
from ._utils import strip_markers

_NO_VALUE = object()

class VariableStack:
    """This represents the variables scope stack of a Compiler program."""

    def __init__(self, stack, executor):
        """Build a new variable stack object with the given stack and program executor."""
        self._stack = stack
        self._executor = executor

    def push(self, variables):
        self._stack.append(variables)

    def pop(self):
        out = self._stack.pop()

        # if we are popping a _prefix variable state we need to update the display
        if "@raw_prefix" in self._stack[-1]:
            self._executor.program.update_display()
        
        return out

    def __getitem__(self, key):
        return self.get(key)
    
    def get(self, name, default_value=KeyError):

        # prefix is a special variable that returns the current prefix without the marker tags
        if name == "@prefix":
            return strip_markers(self.get("@raw_prefix", ""))

        parts = re.split(r"\.|\[", name)
        for variables in reversed(self._stack):
            curr_pos = variables
            found = True
            for part in parts:
                if part.endswith("]"):
                    if re.match(r"['\"0-9].*", part):
                        var_part = ast.literal_eval(part[:-1])
                    else:
                        var_part = self.get(part[:-1])
                else:
                    var_part = part
                try:

                    # check for special computed properties of string values
                    if isinstance(curr_pos, str) and var_part == "__name__":
                        next_pos = self["extract_function_call"](curr_pos).__name__
                    elif isinstance(curr_pos, str) and var_part == "__kwdefaults__":
                        next_pos = self["extract_function_call"](curr_pos).__kwdefaults__
                    else:
                        if isinstance(var_part, str) and hasattr(curr_pos, var_part):
                            next_pos = getattr(curr_pos, var_part)
                        else:
                            next_pos = curr_pos[var_part]
                    next_found = True
                except (KeyError, AttributeError, TypeError):
                    next_found = False
                if next_found:
                    curr_pos = next_pos
                else:
                    found = False
                    break
            if found:
                return curr_pos
        
        # fall back to pulling from the llm namespace
        if not name.startswith("llm."):
            return self.get("llm." + name, default_value)
        
        if default_value is KeyError:
            raise KeyError("`" + name + "` was not found in the program's variables!")
        return default_value # variable not found

    def __contains__(self, name):
        return self.get(name, _NO_VALUE) != _NO_VALUE
    
    def __delitem__(self, key):
        """Note this only works for simple variables, not nested variables."""
        found = True
        for variables in reversed(self._stack):
            if key in variables:
                del variables[key]
                found = True
                break
        if not found:
            raise KeyError(key)

    def __setitem__(self, key, value):
        parts = re.split(r"\.|\[", key)
        found = True
        changed = True
        for variables in reversed(self._stack):
            curr_pos = variables
            found = True
            for part in parts:
                if part.endswith("]"):
                    if re.match(r"[-'\"0-9].*", part):
                        var_part = ast.literal_eval(part[:-1])
                    else:
                        var_part = self.get(part[:-1])
                else:
                    var_part = part
                try:
                    next_pos = curr_pos[var_part]
                    next_found = True
                except KeyError:
                    next_found = False
                
                if next_found:
                    if part == parts[-1]:
                        changed = curr_pos[var_part] != value
                        curr_pos[var_part] = value
                        break
                    else:
                        curr_pos = next_pos
                else:
                    if part == parts[-1] and len(parts) > 1: # setting a new property
                        curr_pos[var_part] = value
                    else:
                        found = False
                    break
            if found:
                break
        if not found:
            assert len(parts) == 1, "Can't set a property of a non-existing variable: " + key
            self._stack[0][key] = value
        
        # if we changed the _prefix variable, update the display
        if changed and key == "@raw_prefix" and not self.get("@no_display", None):
            self._executor.program.update_display()

    def copy(self):
        return VariableStack(self._stack.copy(), self._executor)


from typing import List, Dict, Any
from pydantic import BaseModel

from dotagent import compiler
from dotagent.memory import BaseMemory
from dotagent.schema import BaseMessage
from .prompt import SUMMARIZER_TEMPLATE

def extract_text(string):
    """function for getting the user prompt and llm response from the compiler output"""
    end_string="<|im_end|>"
    start_string="New summary:<|im_end|>\n\n<|im_start|>assistant\n"
    string = string.replace(start_string, "", 1)
    start_index = string.find(start_string)
    end_index = string.find(end_string, start_index + len(start_string))
    return string[start_index + len(start_string):end_index]


class SummaryMemory(BaseMemory, BaseModel):
    current_summary: str = ""
    messages_in_summary: List[Dict[BaseMessage, Any]] = []

    def add_memory(self, prompt: str, llm_response: Any) -> None:
        """Add a self-created message to the store"""
        unique = True
        for conversation in self.messages:
            if conversation['prompt'] == prompt and conversation['llm_response']  == llm_response:
                unique = False
        if unique:
            self.messages.append({'prompt': prompt, 'llm_response': llm_response})

    def get_memory(self, **kwargs) -> str:
        """Retrieve entire memory from the store."""
        
        # Create llm instance
        llm = compiler.llms.OpenAI(model="gpt-3.5-turbo")

        new_messages = [item for item in self.messages if item not in self.messages_in_summary]
        if len(new_messages) != 0:
            messages_to_text=""
            for conversation in new_messages:
                messages_to_text = messages_to_text + "Human: " + conversation['prompt'] + "\n" + "AI: " + conversation['llm_response'] + "\n"
                self.messages_in_summary.append(conversation)

            summarizer = compiler(template=SUMMARIZER_TEMPLATE, llm=llm, stream = False)
            summarized_memory = summarizer(summary=self.current_summary, new_lines= messages_to_text)
            self.current_summary = extract_text(summarized_memory.text)
            summarized_memory = "Current conversation:\n"+self.current_summary

        else:
            summarized_memory=self.current_summary

        return summarized_memory

    def remove_memory(self, prompt: str, llm = Any) -> None:
        """Remove a memory from the store."""

        for conversation in self.messages:
            if conversation['prompt'] == prompt:
                # update messages
                self.messages.remove(conversation)

                # update current summary
                if len(self.messages) != 0:
                    messages_to_text=""
                    self.messages_in_summary.clear()
                    for conversation in self.messages:
                        messages_to_text = messages_to_text + "Human: " + conversation['prompt'] + "\n" + "AI: " + conversation['llm_response'] + "\n"
                        self.messages_in_summary.append(conversation)

                    summarizer = compiler(template=SUMMARIZER_TEMPLATE, llm=llm, stream = False)
                    summarized_memory = summarizer(summary="", new_lines= messages_to_text)
                    summarized_memory = "Current conversation:\n"+extract_text(summarized_memory.text)
                    self.current_summary = summarized_memory
                
                else:
                    self.current_summary=""
                    
                break
    
    def clear(self) -> None:
        """Clear all memories."""
        self.messages.clear()
        self.messages_in_summary.clear()
        self.current_summary = ""


import yaml

from scholarly import scholarly
from scholarly import ProxyGenerator
import os

from dotagent.fact_check.env_config import factool_env_config

# env
# scraper_api_key = factool_env_config.scraper_api_key

class google_scholar():
    def __init__(self):
        pg = ProxyGenerator()
        scraper_api_key = os.environ.get("SCRAPER_API_KEY", None)
        assert scraper_api_key is not None, "Please set the SCRAPER_API_KEY environment variable."
        assert scraper_api_key != '', "Please set the SCRAPER_API_KEY environment variable."
        success = pg.ScraperAPI(scraper_api_key)
        scholarly.use_proxy(pg)

    def run(self, query):
        try:
            results = scholarly.search_pubs(query)
            paper_info = next(results)
            paper_info_subset = {key: paper_info['bib'][key] for key in ['title', 'author', 'pub_year']}
            return paper_info_subset
        except StopIteration:
            return {'title': "no match!", "author": "no match!", "pub_year": "no match!"}



from enum import Enum
from typing import List, Dict, Any, Union, Optional
import os
from dotagent import compiler
from dotagent.tools.basetool import BaseTool
import logging
import yaml
import json
import importlib
import argparse
from dotagent.knowledgebase.doc_loader import document_loader


def import_class(class_path):
    module_name, class_name = class_path.rsplit(".", 1)
    module = importlib.import_module(module_name)
    return getattr(module, class_name)


class AgentState(Enum):
    """Enum to represent different states of the Agent"""

    IDLE = 0
    BUSY = 1
    USED_AS_TOOL = 2
    ERROR = 3


class BaseAgent:
    def __init__(
        self,
        knowledgebase: Optional[Any] = None,
        tools: Optional[Any] = None,
        llm: Optional[Any] = None,
        prompt_template: str = None,
        input_variables: Dict = {},
        agent_id: str = "default",
        memory: Any = None,
        caching: bool = False,
        output_key: str = None,
        return_complete: bool = False,
    ):
        self.agent_id = agent_id
        self.knowledgebase = knowledgebase
        self.tools = tools
        self.llm = llm
        self.prompt_template = prompt_template
        self.input_variables = input_variables
        self.memory = memory
        self.state = AgentState.IDLE
        self.caching = caching
        self.return_complete = return_complete
        self.llm = llm if llm is not None else self.llm_instance()
        self.output_key = output_key

        self.compiler = compiler(
            llm=self.llm, template=self.prompt_template, caching=caching
        )

        if self.return_complete and self.output_key is not None:
            logging.warning(
                "return_complete mode is enabled. Output key will be ignored."
            )

        if self.knowledgebase is not None and not self.input_variables.get(
            "knowledge_variable"
        ):
            raise ValueError(
                "knowledge_variable should be present in input_variables while using knowledge"
            )

    @property
    def agent_type(self):
        pass

    @property
    def get_knowledge_variable(self):
        """Get knowledge variable name from input variables"""
        pass

    @property
    def default_llm_model(self):
        pass

    def add_tool(self, tool: BaseTool) -> None:
        """Add a tool to the agent's tool list."""
        self.tools.append(tool)

    def remove_tool(self, tool: BaseTool) -> None:
        """Remove a tool from the agent's tool list."""
        if tool in self.tools:
            self.tools.remove(tool)

    def llm_instance(self) -> compiler.llms.OpenAI:
        """Create an instance of the language model."""
        pass

    def get_output_key(self, prompt):
        vars = prompt.variables()
        vars = list(vars.keys())
        return vars[-1]

    def get_knowledge(self, query) -> List[str]:
        docs = self.knowledgebase.retrieve_data(query)
        final_doc = ""
        for doc in docs:
            final_doc += doc
        return final_doc

    def run(self, **kwargs) -> Union[str, Dict[str, Any]]:
        """Run the agent to generate a response to the user query."""

        _knowledge_variable = self.get_knowledge_variable

        if _knowledge_variable:
            if kwargs.get(_knowledge_variable):
                query = kwargs.get(_knowledge_variable)
                retrieved_knowledge = self.get_knowledge(query)
                output = self.compiler(
                    RETRIEVED_KNOWLEDGE=retrieved_knowledge, **kwargs, silent=True
                )
            else:
                raise ValueError("knowledge_variable not found in input kwargs")
        else:
            output = self.compiler(**kwargs, silent=True)

        if self.return_complete:
            return output

        _output_key = (
            self.output_key
            if self.output_key is not None
            else self.get_output_key(output)
        )

        if output.variables().get(_output_key):
            return output[_output_key]
        else:
            logging.warning("Output key not found in output, so full output returned")
            return output

    def cli(self):
        """Start a CLI for interacting with the agent."""
        print("Welcome to the dotagent CLI!")

        _vars = []
        for _, v in self.input_variables.items():
            if isinstance(v, str):
                _vars.append(v)
            if isinstance(v, List):
                _vars.extend(v)

        parser = argparse.ArgumentParser(
            description="CLI for interacting with the Agent instance."
        )

        for var in _vars:
            parser.add_argument(f"--{var}", help=f"Pass {var} as an input variable")

        args = parser.parse_args()

        print(self.run(**vars(args)))

    def cli_inputs(self):
        """Start a CLI for interacting with the agent."""
        print("Welcome to the agent CLI!")

        _vars = []
        kwargs = {}
        for _, v in self.input_variables.items():
            if isinstance(v, str):
                _vars.append(v)
            if isinstance(v, List):
                _vars.extend(v)

        for var in _vars:
            kwargs[var] = input(f"{var}: ".capitalize())

        print(self.run(**kwargs))

    def export_agent_config(self, config_path, export_json=False):
        if not config_path.endswith(".agent"):
            return TypeError("config file must have '.agent' extension")

        config = {
            "llm": {
                "type": self.llm.__class__.__module__
                + "."
                + self.llm.__class__.__name__,
                "model": self.llm.model_name,
            },
            "knowledgebase": None
            if self.knowledgebase is None
            else {
                "type": self.knowledgebase.__class__.__module__
                + "."
                + self.knowledgebase.__class__.__name__,
                "data_references": self.knowledgebase.references,
                "data_transformer": {
                    "type": self.knowledgebase.data_transformer.__class__.__module__
                    + "."
                    + self.knowledgebase.data_transformer.__class__.__name__,
                    "chunk_overlap": self.knowledgebase.data_transformer._chunk_overlap,
                    "chunk_size": self.knowledgebase.data_transformer._chunk_size,
                },
                "vector_store": {
                    "type": self.knowledgebase.vector_store.__class__.__module__
                    + "."
                    + self.knowledgebase.vector_store.__class__.__name__,
                    "embedding_function": {
                        "type": self.knowledgebase.vector_store._embedding_function.__class__.__module__
                        + "."
                        + self.knowledgebase.vector_store._embedding_function.__class__.__name__
                    },
                },
            },
            "memory": None
            if self.memory is None
            else {
                "type": self.memory.__class__.__module__
                + "."
                + self.memory.__class__.__name__,
            },
            "prompt_template": self.prompt_template,
            "input_variables": self.input_variables,
            "output_key": self.output_key,
            # 'tools': None if self.tools is None else self.tools
        }
        with open(config_path, "w") as f:
            yaml.dump(config, f)

        if export_json:
            json_path = config_path.replace(config_path.split(".")[-1], "json")
            with open(json_path, "w") as f:
                json.dump(config, f)

    @classmethod
    def load_from_config(cls, config_file):
        with open(config_file, "r") as f:
            config = yaml.safe_load(f)

        # Assume these classes are defined elsewhere and can be imported
        llm_module_name, llm_class_name = config["llm"]["type"].rsplit(".", 1)
        llm_module = importlib.import_module(llm_module_name)
        llm_class = getattr(llm_module, llm_class_name)
        llm = llm_class(model=config["llm"]["model"])

        knowledgebase = None
        if config["knowledgebase"] is not None:
            knowledgebase_module_name, knowledgebase_class_name = config[
                "knowledgebase"
            ]["type"].rsplit(".", 1)
            knowledgebase_module = importlib.import_module(knowledgebase_module_name)
            knowledgebase_class = getattr(
                knowledgebase_module, knowledgebase_class_name
            )

            raw_data = []
            for refer in config["knowledgebase"]["data_references"]:
                reader_class = document_loader(refer["loader_key"])
                raw_docs = reader_class(refer["source"]).load_data()
                raw_data.extend(raw_docs)

            data_transformer_module_name, data_transformer_class_name = config[
                "knowledgebase"
            ]["data_transformer"]["type"].rsplit(".", 1)
            data_transformer_module = importlib.import_module(
                data_transformer_module_name
            )
            data_transformer_class = getattr(
                data_transformer_module, data_transformer_class_name
            )
            chunk_overlap = config["knowledgebase"]["data_transformer"]["chunk_overlap"]
            chunk_size = config["knowledgebase"]["data_transformer"]["chunk_size"]
            data_transformer = data_transformer_class(
                chunk_overlap=chunk_overlap, chunk_size=chunk_size
            )

            embedding_function_module_name, embedding_function_class_name = config[
                "knowledgebase"
            ]["vector_store"]["embedding_function"]["type"].rsplit(".", 1)
            embedding_function_module = importlib.import_module(
                embedding_function_module_name
            )
            embedding_function_class = getattr(
                embedding_function_module, embedding_function_class_name
            )
            embedding_function = embedding_function_class()

            vector_store_module_name, vector_store_class_name = config["knowledgebase"][
                "vector_store"
            ]["type"].rsplit(".", 1)
            vector_store_module = importlib.import_module(vector_store_module_name)
            vector_store_class = getattr(vector_store_module, vector_store_class_name)
            vector_store = vector_store_class(embedding_function=embedding_function)

            knowledgebase = knowledgebase_class(
                raw_data=raw_data,
                data_transformer=data_transformer,
                vector_store=vector_store,
            )

        memory = None
        if config["memory"] is not None:
            # Instantiate the memory here in a similar manner
            memory_module_name, memory_class_name = config["memory"]["type"].rsplit(
                ".", 1
            )
            memory_module = importlib.import_module(memory_module_name)
            memory_class = getattr(memory_module, memory_class_name)
            memory = memory_class()

        agent = cls(
            llm=llm,
            knowledgebase=knowledgebase,
            memory=memory,
            prompt_template=config["prompt_template"],
            input_variables=config["input_variables"],
            output_key=config["output_key"],
        )

        return agent



import asyncio
import copy
import pdb

from dotagent.fact_check.knowledge_qa.pipeline import knowledge_qa_pipeline
from dotagent.fact_check.code.pipeline import code_pipeline
from dotagent.fact_check.math.pipeline import math_pipeline
from dotagent.fact_check.scientific.pipeline import scientific_pipeline
from dotagent.fact_check.med_doc_qa.pipeline import med_doc_qa_pipeline

class Factool():
    def __init__(self, foundation_model):
        self.foundation_model = foundation_model
        self.pipelines = {
                            "kbqa_online": knowledge_qa_pipeline(
                                foundation_model, 10, "online"
                            ),
                            "code": code_pipeline(
                                foundation_model, 3, 3
                            ),
                            "math": math_pipeline(
                                foundation_model
                            ),
                            "scientific": scientific_pipeline(
                                foundation_model
                            ),
                            "med_doc_qa": med_doc_qa_pipeline(
                                foundation_model
                            )
                        }

    def run(self, inputs):
        outputs = copy.deepcopy(inputs)
        batches = []
        current_category = inputs[0]['category']
        current_search_type = inputs[0].get('search_type', None)
        current_data_link = inputs[0].get('data_link', None)
        current_embedding_link = inputs[0].get('embedding_link', None)
        current_batch = []

        for input in inputs:
            if (input['category'] == current_category != 'kbqa') \
                or (input['category'] == current_category == 'kbqa' and input.get('search_type', None) == current_search_type == "online") \
                    or (input['category'] == current_category == 'kbqa' and input.get('search_type', None) == current_search_type == "local"\
                        and input.get('data_link', None)==current_data_link and input.get('embedding_link', None)==current_embedding_link):
                current_batch.append(input)
            else:
                batches.append(current_batch)
                current_batch = [input]
                current_category = input['category']
                current_search_type = input.get('search_type', None)
                current_data_link = input.get('data_link', None)
                current_embedding_link = input.get('embedding_link', None)
                    
        batches.append(current_batch)  # append the last batch

        index = 0
        for batch in batches:
            if not batch: continue
            category = batch[0]['category']
            search_type = batch[0].get('search_type', None)
            if category == 'code':
                batch_results = asyncio.run(
                    self.pipelines[category].run_with_tool_api_call(
                        [sample['prompt'] for sample in batch],
                        [sample['response'] for sample in batch],
                        [sample['entry_point'] for sample in batch]
                    )
                )
            elif category == 'kbqa':
                if search_type is None or search_type == "online":
                    batch_results = asyncio.run(
                        self.pipelines[category+"_online"].run_with_tool_api_call(
                            [sample['prompt'] for sample in batch],
                            [sample['response'] for sample in batch],
                        )
                    )
                else:
                    batch_results = asyncio.run(
                        knowledge_qa_pipeline(
                            self.foundation_model,2,"local",batch[0].get("data_link"),batch[0].get("embedding_link")
                        ).run_with_tool_api_call(
                            [sample['prompt'] for sample in batch],
                            [sample['response'] for sample in batch],
                        )
                    )
            else:
                batch_results = asyncio.run(
                    self.pipelines[category].run_with_tool_api_call(
                        [sample['prompt'] for sample in batch],
                        [sample['response'] for sample in batch]
                    )
                )
            for result in batch_results:
                outputs[index].update(result)
                index += 1
        
        # calculate average response_level_factuality
        total_response_factuality = sum(output['response_level_factuality'] == True for output in outputs)
        avg_response_level_factuality = total_response_factuality / len(outputs)

        # calculate average claim_level_factuality
        num_claims = 0
        total_claim_factuality = 0
        for output in outputs:
            num_claims += len(output['claim_level_factuality'])
            if output['category'] == 'kbqa':
                total_claim_factuality += sum(claim['factuality'] == True for claim in output['claim_level_factuality'])
            elif output['category'] == 'code':
                total_claim_factuality += (output['claim_level_factuality'] == True)
            elif output['category'] == 'math':
                total_claim_factuality += sum(claim_factuality == True for claim_factuality in output['claim_level_factuality'])
            elif output['category'] == 'scientific':
                total_claim_factuality += sum(claim['factuality'] == True for claim in output['claim_level_factuality'])

        avg_claim_level_factuality = total_claim_factuality / num_claims

        return {"average_claim_level_factuality": avg_claim_level_factuality, "average_response_level_factuality": avg_response_level_factuality, "detailed_information": outputs}

    async def run_for_plugin(self, inputs):
        outputs = copy.deepcopy(inputs)

        batches = []
        current_category = inputs[0]['category']
        current_batch = []

        for input in inputs:
            if input['category'] == current_category:
                current_batch.append(input)
            else:
                batches.append(current_batch)
                current_batch = [input]
                current_category = input['category']
                    
        batches.append(current_batch)  # append the last batch

        index = 0
        for batch in batches:
            category = batch[0]['category']
            if category == 'code':
                batch_results = await self.pipelines[category].run_with_tool_api_call(
                    [sample['prompt'] for sample in batch],
                    [sample['response'] for sample in batch],
                    [sample['entry_point'] for sample in batch],
                )
            else:
                batch_results = await self.pipelines[category].run_with_tool_api_call(
                    [sample['prompt'] for sample in batch],
                    [sample['response'] for sample in batch],
                )
            for result in batch_results:
                outputs[index].update(result)
                index += 1
        
        return outputs


import os

from dotagent.audio.speech_to_text.base import SpeechToText


def get_speech_to_text() -> SpeechToText:
    use = os.getenv('SPEECH_TO_TEXT_USE', 'LOCAL_WHISPER')
    if use == 'GOOGLE':
        from dotagent.audio.speech_to_text.google import Google
        Google.initialize()
        return Google.get_instance()
    elif use == 'LOCAL_WHISPER':
        from dotagent.audio.speech_to_text.whisper import Whisper
        Whisper.initialize(use='local')
        return Whisper.get_instance()
    elif use == 'OPENAI_WHISPER':
        from dotagent.audio.speech_to_text.whisper import Whisper
        Whisper.initialize(use='api')
        return Whisper.get_instance()
    else:
        raise NotImplementedError(f'Unknown speech to text engine: {use}')



from typing import Any, List

from dotagent.memory.base import BaseMemory

class ReadOnlyMemory(BaseMemory):
    """Read-only memory class wrapper that prevents modification."""
    memory: BaseMemory

    def __init__(self, memory: BaseMemory):
        self.memory = memory

    def add_memory(self, prompt: str, llm_response: Any) -> None:
        """cannot edit a read only memory."""
        pass

    def get_memory(self, **kwargs) -> Any:
        """Retrieve entire memory from the store."""
        return self.memory.get_memory(**kwargs)

    def remove_memory(self, prompt: str) -> None:
        """cannot edit a read only memory."""
        pass

    def clear(self) -> None:
        """cannot edit a read only memory."""
        pass

    @property
    def memory_keys(self) -> List[str]:
        """Return the prompts for all memories."""
        prompts = []
        for conversation in self.memory.messages:
            prompts.append(conversation['prompt'])
        
        return prompts


import os
import openai
import json
import time
from typing import List, Optional, Union
from pathlib import Path
import csv
from logging import Logger
import logging
from . import LLMFinetune

# openai.organization = "YOUR_ORG_ID"
# APIKEY
# openai.Model.list()


class OpenaiFinetune(LLMFinetune):
    def __init__(self, logger: Logger, openai_key :str ):
        self.logger = logger
        openai.api_key=openai_key

    def transform_data(self, train_csv_file: str, val_csv_file: str, train_output_file: str, val_output_file: str, llm_model: str = "openai") -> str:
        """
        Transforms CSV files into JSONL and creates files for fine-tuning.
        """

        # Verify llm_model
        if llm_model != "openai":
            raise ValueError("Unsupported model:", llm_model)

        # Paths and Output files
        paths = [train_csv_file, val_csv_file]
        output_files = [train_output_file, val_output_file]

        # Extracting prompt-completion pairs
        prompt_completion_pairs = []
        for csv_file in paths:
            with open(csv_file, "r") as f:
                reader = csv.reader(f)
                for row in reader:
                    if len(row) >= 2:
                        prompt = row[0]
                        completion = row[1]
                        prompt_completion_pairs.append((prompt, completion))

        # Writing to JSONL
        for output_file, pairs in zip(output_files, prompt_completion_pairs):
            with open(output_file, "w") as f:
                for pair in pairs:
                    json_obj = {"prompt": pair[0], "completion": pair[1]}
                    json_str = json.dumps(json_obj)
                    f.write(json_str + "\n")

        # Creating Files
        ids = []
        for output_file in output_files:
            if not output_file.endswith(".jsonl"):
                raise Exception(
                    "args `output_file` must be the **file** path to the .jsonl file"
                )
            try:
                _ = openai.File.create(file=open(output_file, "rb"), purpose="fine-tune")
                ids.append(_)
            except Exception as e:
                self.logger.error(f"Error creating file: {e}")
                raise e

        return output_files, ids

    # TODO: Specify use of the method
    # def model(
    #     self,
    #     model_name: str,
    #     input: str,
    #     instruction: str,
    #     n: int,
    #     temperature: float,
    #     top_p: float,
    # ):
    #     try:
    #         model = openai.Edit.create(
    #             model=model_name,
    #             temperature=temperature,
    #             top_p=top_p,
    #             input=input,
    #             instruction=instruction,
    #             n=n,
    #         )
    #         return model
    #     except Exception as e:
    #         self.logger.error(f"Error creating model: {e}")
    #         raise e

    def finetune(
        self,
        training_file: str,
        model_name: Optional[str] = "curie",
        n_epoch: Optional[int] = 4,
        validation_file: Optional[str] = None,
        batch_size: Optional[int] = None,
        learning_rate_multiplier: Optional[int] = None,
        prompt_loss_weight: Optional[int] = 0.01,
        compute_classification_metrics: Optional[bool] = False,
        classification_n_classes: Optional[int] = None,
        classification_positive_class: Optional[str] = None,
        classification_betas: Optional[List[float]] = None,
        suffix: Optional[str] = None,
    ):
        """
        _summary_

        Args:
            training_file (str): The ID of an uploaded file that contains training data.
            model_name (Optional[str], optional): The name of the base model to fine-tune. You can select one of "ada", "babbage", "curie", "davinci", or a fine-tuned model created after 2022-04-21. Defaults to "curie".
            n_epoch (Optional[int], optional):  Number of epochs to train the model for. Defaults to 4.
            validation_file (Optional[str], optional): The ID of an uploaded file that contains validation data. Defaults to None.
            batch_size (Optional[int], optional): Batch size to use for training. Defaults to None.
            learning_rate_multiplier (Optional[int], optional): Learning rate multiplier to use for training. Defaults to None.
            prompt_loss_weight (Optional[int], optional): Weight to use for loss on the prompt tokens. Defaults to 0.01.
            compute_classification_metrics (Optional[bool], optional): If True, classification metrics such as accuracy and f1-score are computed for validation set. Defaults to False.
            classification_n_classes (Optional[int], optional): Number of classes in a classification task. Defaults to None.
            classification_positive_class (Optional[str], optional): This parameter is needed to generate precision, recall, and F1 metrics when doing binary classification. Defaults to None.
            classification_betas (Optional[List[float]], optional): If this is provided, we calculate F-beta scores at the specified beta values. Defaults to None.
            suffix (Optional[str], optional): A string of up to 40 characters that will be added to your fine-tuned model name. Defaults to None.

        Raises:
            e: Errors generated while creating fine-tune job
            Exception: If fine-tuning job fails

        Returns:
            _type_: _description_
        """
        # openai.FineTune.create(training_file="file-XGinujblHPwGLSztz8cPS8XY")

        job_id = None
        try:
            job_id = openai.FineTune.create(
                training_file=training_file,
                model=model_name,
                n_epochs=n_epoch,
                validation_file=validation_file,
                batch_size=batch_size,
                learning_rate_multiplier=learning_rate_multiplier,
                prompt_loss_weight=prompt_loss_weight,
                compute_classification_metrics=compute_classification_metrics,
                classification_n_classes=classification_n_classes,
                classification_positive_class=classification_positive_class,
                classification_betas=classification_betas,
                suffix=suffix,
            )
            while openai.FineTune.retrieve(job_id.get('id')).get('status') == "pending":
                time.sleep(1)
                self.logger.info(
                    "Fine-tuning job status: %s", openai.FineTune.retrieve(job_id.get('id')).get('status')
                )

            if openai.FineTune.retrieve(job_id.get('id')).get('status') == "failed":
                self.logger.error("Fine-tuning job failed")
                raise Exception("Fine-tuning job failed")

            self.logger.info("Fine-tuning job completed successfully")
            return job_id

        except Exception as e:
            self.logger.error(f"Error creating fine-tune job: {e}")
            raise e

if __name__ == "__main__":
    from creds import OPENAI_KEY
    logger = logging.getLogger(__name__)
    logger.setLevel(logging.DEBUG)
    logger.addHandler(logging.StreamHandler())
    finetune = Finetune(logger, openai_key=OPENAI_KEY)
    train_path, val_path = finetune.generate_jsonl_from_csv('sports_train.csv', 'sports_val.csv', 'sports_train.jsonl', 'sports_val.jsonl')
    output_paths, ids = finetune.create_file(output_files=[train_path, val_path])
    train_file, val_file = output_paths
    train_id, val_id = ids
    job_id = finetune.finetune(training_file=train_id.get('id'), n_epoch=1, validation_file=val_id.get('id'), suffix="sports", batch_size=4, compute_classification_metrics=True, classification_n_classes=2, classification_positive_class="hockey", classification_betas=[0.5, 1, 2], prompt_loss_weight=0.01, model_name="curie", learning_rate_multiplier=1.0)
    print("#"*5, end="\n\n")
    print(type(openai.FineTune.retrieve(job_id.get('id'))))
    print(openai.FineTune.retrieve(job_id.get('id')))




from ._llama import LLaMA

class Vicuna(LLaMA):
    """ A HuggingFace transformers version of the Vicuna language model with Compiler support.
    """

    llm_name: str = "vicuna"

    default_system_prompt = """A chat between a curious user and an artificial intelligence assistant. The assistant gives helpful, detailed, and polite answers to the user's questions."""

    @staticmethod
    def role_start(role):
        if role == 'user':
            return 'USER: '
        elif role == 'assistant':
            return 'ASSISTANT: '
        else:
            return ''
    
    @staticmethod
    def role_end(role):
        if role == 'user':
            return ''
        elif role == 'assistant':
            return '</s>'
        else:
            return ''


import pyparsing as pp

pp.ParserElement.enable_packrat()
# pp.enable_diag(pp.Diagnostics.enable_debug_on_named_expressions)
# pp.autoname_elements()

class SavedTextNode:
    """A node that saves the text it matches."""
    def __init__(self, s, loc, tokens):
        start_pos = tokens[0]
        if len(tokens) == 3:
            end_pos = tokens[2]
        else:
            end_pos = loc
        self.text = s[start_pos:end_pos]
        assert len(tokens[1]) == 1
        self.tokens = tokens[1][0]
    def __repr__(self):
        return "SavedTextNode({})".format(self.text) + self.tokens.__repr__()
    def __getitem__(self, item):
        return self.tokens[item]
    def __len__(self):
        return len(self.tokens)
    def get_name(self):
        return self.tokens.get_name()
    def __contains__(self, item):
        return item in self.tokens
    def __getattr__(self, name):
        return getattr(self.tokens, name)
    def __call__(self, *args, **kwds):
        return self.tokens(*args, **kwds)
def SavedText(node):
    return pp.Located(node).add_parse_action(SavedTextNode)

program = pp.Forward()
program_chunk = pp.Forward()

## whitespace ##

ws = pp.White()
opt_ws = pp.Optional(ws)


## comments ##

# long-form comments {{!-- my comment --}}
command_end = pp.Suppress(opt_ws + "}}") | pp.Suppress(opt_ws + "~}}" + opt_ws)
long_comment_start = pp.Suppress(pp.Literal("{{") + pp.Optional("~") + pp.Literal("!--"))
long_comment_end =  pp.Suppress(pp.Literal("--") + command_end)
not_long_comment_end = "-" + ~pp.FollowedBy("-}}") + ~pp.FollowedBy("-~}}")
long_comment_content = not_long_comment_end | pp.OneOrMore(pp.CharsNotIn("-"))
long_comment = SavedText(pp.Group(pp.Combine(long_comment_start + pp.ZeroOrMore(long_comment_content) + long_comment_end))("long_comment").set_name("long_comment"))

# short-form comments  {{! my comment }}
comment_start = pp.Suppress("{{" + pp.Optional("~") + "!")
not_comment_end = "}" + ~pp.FollowedBy("}") | "~" + ~pp.FollowedBy("}}")
comment_content = not_comment_end | pp.OneOrMore(pp.CharsNotIn("~}"))
comment = SavedText(pp.Group(pp.Combine(comment_start + pp.ZeroOrMore(comment_content) + command_end))("comment"))


## literals ##

literal = pp.Forward().set_name("literal")

# basic literals
string_literal = pp.Group(pp.Suppress('"') + pp.ZeroOrMore(pp.CharsNotIn('"')) + pp.Suppress('"') | pp.Suppress("'") + pp.ZeroOrMore(pp.CharsNotIn("'")) + pp.Suppress("'"))("string_literal")
number_literal = pp.Group(pp.Word(pp.srange("[0-9.]")))("number_literal")
boolean_literal = pp.Group("True" | pp.Literal("False"))("boolean_literal")

# object literal
object_literal = pp.Forward().set_name("object_literal")
object_start = pp.Suppress("{")
object_end = pp.Suppress("}")
empty_object = object_start + object_end
object_item = string_literal + pp.Suppress(":") + literal
single_item_object = object_start + object_item + object_end
object_sep = pp.Suppress(",")
multi_item_object = object_start + object_item + pp.ZeroOrMore(object_sep + object_item) + object_end
object_literal <<= pp.Group(empty_object | single_item_object | multi_item_object)("object_literal")

# array literal
array_literal = pp.Forward().set_name("array_literal")
array_start = pp.Suppress("[")
array_end = pp.Suppress("]")
array_item = literal
empty_array = array_start + array_end
single_item_array = array_start + array_item + array_end
array_sep = pp.Suppress(",")
multi_item_array = array_start + array_item + pp.ZeroOrMore(array_sep + array_item) + array_end
array_literal <<= pp.Group(empty_array | single_item_array | multi_item_array)("array_literal")

literal <<= string_literal | number_literal | boolean_literal | array_literal | object_literal


## infix operators ##

code_chunk_no_infix = pp.Forward().set_name("code_chunk_no_infix")

class OpNode:
    def __repr__(self):
        return "{}({})".format(self.__class__.__name__, self.operator)
    def __getitem__(self, item):
        return getattr(self, item)
    def get_name(self):
        return self.name

class UnOp(OpNode):
    def __init__(self, tokens):
        self.operator = tokens[0][0]
        self.value = tokens[0][1]
        self.name = "unary_operator"

class BinOp(OpNode):
    def __init__(self, tokens):
        self.operator = tokens[0][1]
        self.lhs = tokens[0][0]
        self.rhs = tokens[0][2]
        self.name = "binary_operator"

infix_operator_block = pp.infix_notation(code_chunk_no_infix, [
    (pp.one_of('- not'), 1, pp.OpAssoc.RIGHT, UnOp),
    (pp.one_of('* /'), 2, pp.OpAssoc.LEFT, BinOp),
    (pp.one_of('+ -'), 2, pp.OpAssoc.LEFT, BinOp),
    (pp.one_of('< > <= >= == != is in'), 2, pp.OpAssoc.LEFT, BinOp),
    (pp.one_of('and'), 2, pp.OpAssoc.LEFT, BinOp),
    (pp.one_of('or'), 2, pp.OpAssoc.LEFT, BinOp),
])


## commands ##

code_chunk = pp.Forward().set_name("code_chunk")
not_keyword = ~pp.FollowedBy(pp.Keyword("or") | pp.Keyword("else") | pp.Keyword("elif") | pp.Keyword("not"))
command_name = pp.Combine(not_keyword + pp.Word(pp.srange("[@A-Za-z_]"), pp.srange("[A-Za-z_0-9\.]")))
variable_name = pp.Word(pp.srange("[@A-Za-z_]"), pp.srange("[A-Za-z_0-9]"))
variable_ref = not_keyword + pp.Group(pp.Word(pp.srange("[@A-Za-z_]"), pp.srange("[@A-Za-z_0-9\.\[\]\"'-]")))("variable_ref").set_name("variable_ref")
keyword = pp.Group(pp.Keyword("break") | pp.Keyword("continue"))("keyword")

# command arguments
command_arg = pp.Forward()
named_command_arg = variable_name + "=" + code_chunk
command_arg <<= pp.Group(named_command_arg)("named_command_arg").set_name("named_command_arg") | pp.Group(code_chunk)("positional_command_arg").set_name("positional_command_arg")

# whitespace command format {{my_command arg1 arg2=val}}
ws_command_call = pp.Forward().set_name("ws_command_call")
command_arg_and_ws = pp.Suppress(ws) + command_arg
ws_command_args = pp.OneOrMore(command_arg_and_ws)
# note that we have to list out all the operators here because we match before the infix operator checks
ws_command_call <<= command_name("name") + ~pp.FollowedBy(pp.one_of("+ - * / or not is and <= == >= != < >")) + ws_command_args

# paren command format {{my_command(arg1, arg2=val)}}
paren_command_call = pp.Forward().set_name("paren_command_call")
command_arg_and_comma_ws = pp.Suppress(",") + command_arg
paren_command_args = pp.Optional(command_arg) + pp.ZeroOrMore(command_arg_and_comma_ws)
paren_command_call <<= (command_name("name") + pp.Suppress("(")).leave_whitespace() - paren_command_args + pp.Suppress(")")

# code chunks
enclosed_code_chunk = pp.Forward().set_name("enclosed_code_chunk")
paren_group = (pp.Suppress("(") - enclosed_code_chunk + pp.Suppress(")")).set_name("paren_group")
enclosed_code_chunk_cant_infix = (pp.Group(ws_command_call)("command_call") | pp.Group(paren_command_call)("command_call") | literal | keyword | variable_ref | paren_group) + ~pp.FollowedBy(pp.one_of("+ - * / or not is and <= == >= != < >"))
enclosed_code_chunk <<= enclosed_code_chunk_cant_infix | infix_operator_block
code_chunk_no_infix <<= (paren_group | pp.Group(paren_command_call)("command_call") | literal | keyword | variable_ref) # used by infix_operator_block
code_chunk_cant_infix = code_chunk_no_infix + ~pp.FollowedBy(pp.one_of("+ - * / or not is and <= == >= != < >")) # don't match infix operators so we can run this before infix_operator_block
code_chunk_cant_infix.set_name("code_chunk_cant_infix")
code_chunk <<= code_chunk_cant_infix | infix_operator_block

# command/variable
command_start = pp.Suppress("{{" + ~pp.FollowedBy("!") + pp.Optional("~"))
simple_command_start = pp.Suppress("{{" + ~pp.FollowedBy("!") + pp.Optional("~")) + ~pp.FollowedBy(pp.one_of("# / >"))
command = SavedText(pp.Group(simple_command_start + enclosed_code_chunk + command_end)("command"))

# partial
always_call = pp.Group(paren_command_call | command_name("name") + pp.Optional(ws_command_args))
partial = pp.Group(pp.Suppress(pp.Combine(command_start + ">")) + always_call("command_call") + command_end)("partial")

# block command {{#my_command arg1 arg2=val}}...{{/my_command}}
separator = pp.Group(pp.Keyword("or") | pp.Keyword("else") | (pp.Keyword("elif") + ws_command_args))("separator").set_name("separator")
block_command = pp.Forward()
block_command_call = always_call("command_call")
block_command_open = pp.Suppress(pp.Combine(command_start + "#")) + block_command_call + command_end
block_command_sep = (command_start + separator + command_end)("block_command_sep").set_name("block_command_sep")
block_command_close = SavedText(pp.Group(command_start + pp.Suppress("/") + command_name + command_end)("block_command_close").set_name("block_command_close"))
block_command_content = (pp.Group(program)("block_content_chunk") + pp.ZeroOrMore(block_command_sep + pp.Group(program)("block_content_chunk"))).set_name("block_content")
block_command <<= (block_command_open + SavedText(pp.Group(block_command_content)("block_content")) + block_command_close).leave_whitespace()
block_command = SavedText(pp.Group(block_command)("block_command")).set_name("block_command")

# block partial {{#>my_command arg1 arg2=val}}...{{/my_command}}
block_partial = pp.Forward()
block_partial_call = always_call("command_call")
block_partial_open = pp.Combine(command_start + pp.Suppress("#>")) + block_partial_call + command_end
block_partial_close = command_start + pp.Suppress("/") + command_name + command_end
block_partial <<= block_partial_open + program + pp.Suppress(block_partial_close)
block_partial = SavedText(pp.Group(block_partial)("block_partial"))

# escaped commands \{{ not a command }}
not_command_end = "}" + ~pp.FollowedBy("}")
escaped_command = SavedText(pp.Group(pp.Suppress("\\") + command_start + pp.Combine(pp.ZeroOrMore(pp.CharsNotIn("}") | not_command_end)) + command_end)("escaped_command"))
unrelated_escape = "\\" + ~pp.FollowedBy(command_start)


## content ##

not_command_start = "{" + ~pp.FollowedBy("{" + pp.CharsNotIn("{"))
not_command_escape = "\\" + ~pp.FollowedBy("{{")
stripped_whitespace = pp.Suppress(pp.Word(" \t\r\n")) + pp.FollowedBy("{{~")
unstripped_whitespace = pp.Word(" \t\r\n") # no need for a negative FollowedBy because stripped_whitespace will match first
content = pp.Group(pp.Combine(pp.OneOrMore(stripped_whitespace | unstripped_whitespace | not_command_start | not_command_escape | pp.CharsNotIn("{\\ \t\r\n"))))("content").set_name("content")

# keyword_command = SavedText(pp.Group(command_start + keyword + ws_command_args + command_end)("keyword_command"))
# block_content_chunk = long_comment | comment | escaped_command | unrelated_escape | block_partial | block_command | partial | command | content
# block_content <<= pp.ZeroOrMore(block_content_chunk)("program").leave_whitespace()

## global program ##

program_chunk <<= (long_comment | comment | escaped_command | unrelated_escape | block_partial | block_command | partial | command | content).leave_whitespace()
program <<= pp.ZeroOrMore(program_chunk)("program").leave_whitespace().set_name("program")
grammar = (program + pp.StringEnd()).parse_with_tabs()


from typing import List, Dict, Any
from pydantic import BaseModel

from dotagent import compiler
from dotagent.memory import BaseMemory
from dotagent.schema import BaseMessage
from .prompt import SUMMARIZER_TEMPLATE

def extract_text(string):
    """function for getting the user prompt and llm response from the compiler output"""
    end_string="<|im_end|>"
    start_string="New summary:<|im_end|>\n\n<|im_start|>assistant\n"
    string = string.replace(start_string, "", 1)
    start_index = string.find(start_string)
    end_index = string.find(end_string, start_index + len(start_string))
    return string[start_index + len(start_string):end_index]


class BufferSummaryMemory(BaseMemory, BaseModel):
    current_summary: str = ""
    current_buffer: str = ""
    messages_in_summary: List[Dict[BaseMessage, Any]] = []
    messages_in_buffer: List[Dict[BaseMessage, Any]] = []

    def add_memory(self, prompt: str, llm_response: Any) -> None:
        """Add a self-created message to the store"""
        unique = True
        for conversation in self.messages:
            if conversation['prompt'] == prompt and conversation['llm_response']  == llm_response:
                unique = False
        if unique:
            self.messages.append({'prompt': prompt, 'llm_response': llm_response})

    def get_memory(self, **kwargs) -> str:
        """Retrieve entire memory from the store."""

        # Create llm instance
        llm = compiler.llms.OpenAI(model="gpt-3.5-turbo")
        
        # get variables
        if 'memory_threshold' not in kwargs.keys():
            memory_threshold = 1
        else:
            memory_threshold = kwargs['memory_threshold']

        # get new messages in memory
        new_messages = [item for item in self.messages if item not in self.messages_in_summary and item not in self.messages_in_buffer]

        if len(new_messages) != 0:
            # split conversations for buffer and summary according to threshold         
            buffer_messages = self.messages[-1*memory_threshold:] # buffer
            summary_messages =  [item for item in self.messages if item not in buffer_messages] # summary
            
            # get messages in the buffer format as text
            self.current_buffer = ""
            for conversation in buffer_messages:
                self.current_buffer = self.current_buffer + "Human: " + conversation['prompt'] + "\n" + "AI: " + conversation['llm_response'] + "\n"
            self.messages_in_buffer=buffer_messages

            # get the new messages to be added in the summary
            additions_to_summary = [item for item in summary_messages if item not in self.messages_in_summary]

            if len(additions_to_summary) != 0:
                additions_to_summary_as_text=""
                
                for conversation in additions_to_summary:
                    additions_to_summary_as_text = additions_to_summary_as_text + "Human: " + conversation['prompt'] + "\n" + "AI: " + conversation['llm_response'] + "\n"

                self.messages_in_summary = summary_messages

                summarizer = compiler(template=SUMMARIZER_TEMPLATE, llm=llm, stream = False)
                summarized_memory = summarizer(summary=self.current_summary, new_lines= additions_to_summary_as_text)
                self.current_summary = extract_text(summarized_memory.text)
                
                summarized_memory = "Current conversation:\n"+self.current_summary + "\n" + self.current_buffer
            else:
                summarized_memory= "Current conversation:\n"+self.current_buffer
        
        else:
            if len(self.current_buffer) != 0 or len(self.current_summary) != 0: 
                summarized_memory = "Current conversation:\n"+self.current_summary + "\n" + self.current_buffer
            else:
                summarized_memory = ""
                
        return summarized_memory

    def remove_memory(self, prompt: str, llm: Any, threshold=1) -> None:
        """Remove a memory from the store."""

        # update messages
        for conversation in self.messages:
            if conversation['prompt'] == prompt:
                # updating messages
                self.messages.remove(conversation)
                
                # updating other variables
                self.current_summary = ""
                self.current_buffer = ""
                self.messages_in_buffer.clear()
                self.messages_in_summary.clear()

                # get new messages in memory
                new_messages = [item for item in self.messages if item not in self.messages_in_summary and item not in self.messages_in_buffer]

                if len(new_messages) != 0:
                    # split conversations for buffer and summary according to threshold         
                    buffer_messages = self.messages[-1*threshold:] # buffer
                    summary_messages =  [item for item in self.messages if item not in buffer_messages] # summary
                    
                    # get messages in the buffer format as text
                    self.current_buffer = ""
                    for conversation in buffer_messages:
                        self.current_buffer = self.current_buffer + "Human: " + conversation['prompt'] + "\n" + "AI: " + conversation['llm_response'] + "\n"
                    self.messages_in_buffer=buffer_messages

                    # get the new messages to be added in the summary
                    additions_to_summary = [item for item in summary_messages if item not in self.messages_in_summary]

                    if len(additions_to_summary) != 0:
                        additions_to_summary_as_text=""
                        print(additions_to_summary_as_text)
                        
                        for conversation in additions_to_summary:
                            additions_to_summary_as_text = additions_to_summary_as_text + "Human: " + conversation['prompt'] + "\n" + "AI: " + conversation['llm_response'] + "\n"

                        self.messages_in_summary = summary_messages

                        summarizer = compiler(template=SUMMARIZER_TEMPLATE, llm=llm, stream = False)
                        summarized_memory = summarizer(summary=self.current_summary, new_lines= additions_to_summary_as_text)
                        self.current_summary = extract_text(summarized_memory.text)
                break
    
    def clear(self) -> None:
        """Clear all memories."""
        self.messages.clear()
        self.messages_in_summary.clear()
        self.messages_in_buffer.clear()
        self.current_summary = ""
        self.current_buffer = ""


from dotagent.vectorstores.base import VectorStore
from dotagent.vectorstores.chroma import Chroma
from dotagent.vectorstores.deeplake import DeepLake
from dotagent.vectorstores.pinecone import Pinecone
from dotagent.vectorstores.qdrant import Qdrant
from dotagent.vectorstores.redis import Redis




__all__ = ['VectorStore', 'Chroma', 'DeepLake', 'Pinecone', 'Qdrant', 'Redis']



#Code taken from langchain Github Repo and removed load method as it involved Documents from langchain

"""Util that calls Wikipedia."""
import logging
from typing import Any, Dict, List, Optional

from pydantic import BaseModel, Extra, root_validator

logger = logging.getLogger(__name__)

WIKIPEDIA_MAX_QUERY_LENGTH = 300


class WikipediaAPIWrapper(BaseModel):
    """Wrapper around WikipediaAPI.

    To use, you should have the ``wikipedia`` python package installed.
    This wrapper will use the Wikipedia API to conduct searches and
    fetch page summaries. By default, it will return the page summaries
    of the top-k results.
    It limits the Document content by doc_content_chars_max.
    """

    wiki_client: Any  #: :meta private:
    top_k_results: int = 3
    lang: str = "en"
    load_all_available_meta: bool = False
    doc_content_chars_max: int = 4000

    class Config:
        """Configuration for this pydantic object."""

        extra = Extra.forbid

    @root_validator()
    def validate_environment(cls, values: Dict) -> Dict:
        """Validate that the python package exists in environment."""
        try:
            import wikipedia

            wikipedia.set_lang(values["lang"])
            values["wiki_client"] = wikipedia
        except ImportError:
            raise ImportError(
                "Could not import wikipedia python package. "
                "Please install it with `pip install wikipedia`."
            )
        return values

    def run(self, query: str) -> str:
        """Run Wikipedia search and get page summaries."""
        page_titles = self.wiki_client.search(query[:WIKIPEDIA_MAX_QUERY_LENGTH])
        summaries = []
        for page_title in page_titles[: self.top_k_results]:
            if wiki_page := self._fetch_page(page_title):
                if summary := self._formatted_page_summary(page_title, wiki_page):
                    summaries.append(summary)
        if not summaries:
            return "No good Wikipedia Search Result was found"
        return "\n\n".join(summaries)[: self.doc_content_chars_max]

    @staticmethod
    def _formatted_page_summary(page_title: str, wiki_page: Any) -> Optional[str]:
        return f"Page: {page_title}\nSummary: {wiki_page.summary}"


    def _fetch_page(self, page: str) -> Optional[str]:
        try:
            return self.wiki_client.page(title=page, auto_suggest=False)
        except (
            self.wiki_client.exceptions.PageError,
            self.wiki_client.exceptions.DisambiguationError,
        ):
            return None


import traceback
import ast
import inspect
import re
import asyncio
import logging
import pyparsing as pp
from ._grammar import grammar
from ._variable_stack import VariableStack
log = logging.getLogger(__name__)


class ProgramExecutor():
    def __init__(self, program):
        """ Attaches this executor to a program object.
        """

        self.program = program
        self.block_content = []
        self.executing = True
        self.should_stop = False
        self.caught_stop_iteration = False
        self.llm_session = None
        self._logging = hasattr(self.program.log, "append")

        # find all the handlebars-style partial inclusion tags and replace them with the partial template
        # def replace_partial(match):
        #     parts = match.group(1).split(" ", 1)
        #     partial_name = parts[0]
            
        #     # ,args_string = match.group(1).split(" ", 1)
        #     if partial_name not in program._variables:
        #         raise ValueError("Partial '%s' not given in the keyword args:" % partial_name)
        #     out = "{{#block '"+partial_name+"'"
        #     if len(parts) > 1:
        #         out += " " + parts[1]
        #     out += "}}" + program._variables[partial_name].text + "{{/block}}"
        #     # Update the current program variables using those from the partial, but do not overwrite.
        #     # (Rebuilding the _variables map here would break returning new values to the program variables later, e.g. from gen.)
        #     update_variables = {
        #         k: v
        #         for k, v in program[partial_name]._variables.items()
        #         if k not in program._variables
        #     }
        #     program._variables.update(update_variables)
        #     return out
        # text = re.sub(r"{{>(.*?)}}", replace_partial, program._text)

        # parse the program text
        try:
            self.parse_tree = grammar.parse_string(program._text)
        except (pp.ParseException, pp.ParseSyntaxException) as e:
            initial_str = program._text[max(0, e.loc-40):e.loc]
            initial_str = initial_str.split("\n")[-1] # trim off any lines before the error
            next_str = program._text[e.loc:e.loc+40]
            error_string = str(e)
            if next_str.startswith("{{#") or next_str.startswith("{{~#"):
                error_string += "\nPerhaps the block command was not correctly closed?"
            msg = error_string + "\n\n"+initial_str
            # msg += "\033[91m" + program._text[e.loc:e.loc+40] + "\033[0m\n"
            msg += program._text[e.loc:e.loc+40] + "\n"
            msg += " " * len(initial_str) + "^\n"

            raise SyntaxException(msg, e) from None
            
    # def _check_for_simple_error(self, text):
    #     """ Check for a simple errors in the program text, and give nice error messages.
    #     """

    #     vars = self.program._variables

    #     # missing block pound sign
    #     for k in vars:
    #         if getattr(vars[k], "is_block", False):

    #             # look for block commands that are missing the opening pound sign or closing slash
    #             m = re.search(r"(^|[^\\]){{\s*"+k+"(\s|}|~)", text)
    #             if m is not None:
    #                 # get the context around the matching error
    #                 start = max(0, m.start()-30)
    #                 end = min(len(text), m.end()+30)
    #                 context = text[start:end]
    #                 if start > 0:
    #                     context = "..."+context
    #                 if end < len(text):
    #                     context = context+"..."
    #                 raise ValueError("The Compiler program is missing the opening pound (#) sign or closing slash (/) for the block level command `"+k+"` at:\n"+context) from None
                
    #             # look for block commands that are missing the closing tag
    #             num_opens = len(re.findall(r"(^|[^\\]){{~?#\s*"+k+"(\s|}|~)", text))
    #             num_closes = len(re.findall(r"(^|[^\\]){{~?/\s*"+k+"(\s|}|~)", text))
    #             if num_opens > num_closes:
    #                 raise ValueError("The Compiler program is missing a closing tag for the block level command `"+k+"`.") from None
    #             if num_opens < num_closes:
    #                 raise ValueError("The Compiler program is missing an opening tag for the block level command `"+k+"`.") from None
        
        

    async def run(self, llm_session):
        """ Execute the program.
        """
        self.llm_session = llm_session
        try:
            # first parse all the whitespace control
            # self.whitespace_control_visit(self.parse_tree)

            # now execute the program
            self.program._variables["@raw_prefix"] = ""
            await self.visit(self.parse_tree, VariableStack([self.program._variables], self))
        except Exception as e:
            print(traceback.format_exc())
            print("Error in program: ", e)
            raise e
        
    def stop(self):
        self.should_stop = True

    # def process_content(self, text):
    #     if text.endswith("{{!--GSTRIP--}}"):
    #         text = text[:-15].rstrip()
    #     if text.startswith("{{!--GSTRIP--}}"):
    #         text = text[15:].lstrip()
    #     return text

    # def whitespace_control_visit(self, node, next_node=None, prev_node=None, parent_node=None, grandparent_node=None):
    #     if node_name in ('command', 'block_command_open', 'block_command_sep', 'block_command_close'):
    #         if node.text.startswith("{{~"):
    #             if prev_node and prev_node_name == "content":
    #                 prev_node.text = prev_node.text + "{{!--GSTRIP--}}"
    #         if node.text.endswith("~}}"):
    #             if next_node and next_node_name == "content":
    #                 next_node.text = "{{!--GSTRIP--}}" + next_node.text

    #     # visit all our children
    #     for i, child in enumerate(node.children):
    #         if len(node.children) > i + 1:
    #             inner_next_node = node.children[i + 1]
    #         else:
    #             inner_next_node = next_node
    #         if i > 0:
    #             inner_prev_node = node.children[i - 1]
    #         else:
    #             inner_prev_node = prev_node
    #         self.whitespace_control_visit(child, inner_next_node, inner_prev_node, node, parent_node)
    
    async def visit(self, node, variable_stack, next_node=None, next_next_node=None, prev_node=None, parent_node=None, grandparent_node=None):

        # if we are after a break point then we return nothing
        # (note that this flag will be cleared once the loop is ended)
        if self.caught_stop_iteration:
            return ""
        
        node_name = node.get_name()

        if node_name == 'variable_name':
            return node.text

        elif node_name == 'content':
            variable_stack["@raw_prefix"] += node[0]
            return None
        
        elif node_name == 'long_comment':
            if node.text.startswith("{{!--G"):
                variable_stack["@raw_prefix"] += node.text
            return ""
        
        elif node_name == 'comment':
            return ""
        
        elif node_name == 'partial':
            partial_program = variable_stack[node[0]["name"]]
            tree = grammar.parse_string(partial_program._text)
            partial_args = [await self.visit(child, variable_stack) for child in node["command_call"][1:]]
            args = []
            kwargs = {}
            for arg in partial_args:
                if isinstance(arg, PositionalArgument):
                    args.append(arg.value)
                elif isinstance(arg, NamedArgument):
                    kwargs[arg.name] = arg.value
            partial_vars = {k: v for k,v in partial_program.variables().items() if k not in ["llm", "logging"] and k not in variable_stack}
            if len(args) > 0:
                partial_vars["args"] = args
            partial_vars.update(kwargs)
            variable_stack.push(partial_vars)
            out = await self.visit(tree, variable_stack)
            variable_stack.pop()
            return out

        elif node_name == 'positional_command_arg':
            return PositionalArgument(await self.visit(node[0], variable_stack))

        elif node_name == 'named_command_arg':
            return NamedArgument(node[0], await self.visit(node[2], variable_stack))

        elif node_name == 'command_name':
            return node.text
        
        elif node_name == 'command_name':
            return node.text

        elif node_name == 'escaped_command':
            variable_stack["@raw_prefix"] += node.text[1:]
            return
        
        elif node_name == 'boolean_literal':
            if node[0] == "True":
                return True
            elif node[0] == "False":
                return False
            else:
                raise Exception("Invalid boolean literal")

        elif node_name == 'number_literal':
            if "." in node[0]:
                return float(node[0])
            else:
                return int(node[0])
            
        elif node_name == 'string_literal':
            return node[0]
        
        elif node_name == 'object_literal':
            out = {}
            for i in range(0, len(node), 2):
                key = await self.visit(node[i], variable_stack)
                value = await self.visit(node[i + 1], variable_stack)
                out[key] = value
            return out
        
        elif node_name == 'array_literal':
            return [await self.visit(node[i], variable_stack) for i in range(0, len(node))]

        elif node_name == 'literal':
            try:
                return ast.literal_eval(node.text)
            except Exception as e:
                raise Exception(f"Error parsing literal: {node.text} ({e})")

        elif node_name == 'command':

            # if execution is already stopped before we start the command we just keep the command text
            if not self.executing:
                variable_stack["@raw_prefix"] += node.text
                return
            
            # mark our position in case we need to rewind
            # pos = len(self.prefix)

            # find the command name
            if "variable_ref" in node:
                name = "variable_ref"
            elif "keyword" in node:
                name = "keyword"
            elif "command_call" in node:
                name = node["command_call"]["name"]
            else: # binary_operator and unary_operator
                name = node[0].get_name()

            # add the start marker
            escaped_node_text = node.text.replace("$", "&#36;").replace("{", "&#123;").replace("}", "&#125;")
            variable_stack["@raw_prefix"] += "{{!--"+f"GMARKER_START_{name}${escaped_node_text}$"+"--}}"
            
            # visit our children
            self.block_content.append([])
            visited_children = [await self.visit(child, variable_stack, next_node, next_next_node, prev_node, node, parent_node) for child in node]
            self.block_content.pop()
            out = "".join("" if c is None else str(c) for c in visited_children)

            variable_stack["@raw_prefix"] += out +  "{{!--" + f"GMARKER_END_{name}$$" + "--}}"

            # if execution became stopped during the command, we append the command text
            if not self.executing:
                # self.reset_prefix(pos)
                variable_stack["@raw_prefix"] += node.text
            return

        elif node_name == 'paren_group':
            visited_children = [await self.visit(child, variable_stack) for child in node.children]
            return visited_children[1]

        elif node_name == 'command_call' or node_name == 'variable_ref' or node_name == 'binary_operator' or node_name == 'unary_operator' or node_name == 'keyword':
            if node_name == 'command_call':
                command_name = node["name"]
                args = [await self.visit(child, variable_stack) for child in node[1:]]
            elif node_name == 'binary_operator':
                command_name = "BINARY_OPERATOR_" + node["operator"]
                args = [
                    PositionalArgument(await self.visit(node["lhs"], variable_stack)),
                    PositionalArgument(await self.visit(node["rhs"], variable_stack))
                ]
            elif node_name == 'unary_operator':
                command_name = "UNARY_OPERATOR_" + node["operator"]
                args = [PositionalArgument(await self.visit(node["value"], variable_stack))]
            else:
                command_name = node[0]
                args = []

            # if the command arguments stopped execution, we don't execute the command
            if not self.executing:
                return
            
            # return_value = ""
            if command_name in variable_stack:
                command_function = variable_stack[command_name]

                # we convert a variable reference to a function that returns the variable value
                if node_name == "variable_ref":
                    command_value = command_function
                    command_function = lambda: command_value

                # check for a generated call statement
                named_args = {}
                if isinstance(command_function, str):
                    call_details = variable_stack["extract_function_call"](command_function)
                    if call_details is None:
                        raise Exception(f"Can't call the string (there is no function call recognized by `extract_function_call` in it): {command_function}")
                    
                    command_function = call_details.__name__
                    if command_function not in variable_stack:
                        raise Exception(f"Function {command_function} not found!")
                    else:
                        command_function = variable_stack[command_function]
                        named_args = call_details.__kwdefaults__


                # def update_return_value(s):
                #     nonlocal return_value
                #     if return_value == "":
                #         return_value = s
                    
                #     # convert to strings if we are concatenating
                #     else:
                #         return_value += "" if s is None else str(s)

                # If we are a top level command we extend the prefix
                top_level = parent_node is not None and parent_node.get_name() == "command"
                    # partial_output = self.extend_prefix
                    # pass
                
                # otherwise we keep track of output locally so we can return it
                if not top_level:
                    # partial_output = update_return_value
                    pos = len(variable_stack["@raw_prefix"])
                    variable_stack.push({"@raw_prefix": variable_stack["@raw_prefix"], "@no_display": True})

                # create the arguments for the command
                positional_args = []
                for arg in args:
                    if isinstance(arg, PositionalArgument):
                        positional_args.append(arg.value)
                    elif isinstance(arg, NamedArgument):
                        named_args[arg.name] = arg.value
                sig = inspect.signature(command_function)
                if "_parser_context" in sig.parameters:
                    named_args["_parser_context"] = {
                        "parser": self,
                        "variable_stack": variable_stack,
                        "next_node": next_node,
                        "next_next_node": next_next_node,
                        "prev_node": prev_node,
                        "block_content": None
                    }

                # call the command
                if self._logging:
                    self.program.log.append({
                        "type": "start",
                        "name": command_name,
                        "positional_args": positional_args,
                        "named_args": {k:v for k,v in named_args.items() if k != "_parser_context"},
                        "@prefix": variable_stack["@prefix"],
                        # "node_id": id(node)
                    })
                    pos = len(variable_stack["@prefix"])
                try:
                    if inspect.iscoroutinefunction(command_function):
                        await asyncio.sleep(0) # give other coroutines a chance to run
                        command_output = await command_function(*positional_args, **named_args)
                    else:
                        command_output = command_function(*positional_args, **named_args)
                except StopIteration as ret:
                    command_output = ret.value
                    self.caught_stop_iteration = True
                if self._logging:
                    self.program.log.append({"type": "end", "name": command_name, "new_prefix": variable_stack["@prefix"][pos:]})

                # call partial output if the command didn't itself (and we are still executing)
                if not top_level:
                    curr_prefix = variable_stack.pop()["@raw_prefix"] # pop the variable stack we pushed earlier becuause we were hidden
                    if command_output is not None:
                        return command_output
                    else:
                        new_content = curr_prefix[pos:]

                        # see if we got a list of outputs encoded as a string
                        parts = re.split(r"{{!--GMARKERmany[^}]+}}", new_content)
                        if len(parts) > 1:
                            return parts[1:-1]
                        else:
                            return new_content
                else:
                    if command_output is not None:
                        variable_stack["@raw_prefix"] += str(command_output)
                    return ""
            else:
                # if the variable does not exist we just pause execution
                if self.program.await_missing:
                    self.executing = False
                    return None
                else:
                    # raise an error if the command doesn't exist
                    raise KeyError("Command/variable '"+command_name+"' not found! Please pass it when calling the program (or set a default value for it when creating the program).")
            
            # # if we are not a top level command we return the output instead of displaying it
            # if not top_level:
            #     return_value = variable_stack.pop()["@raw_prefix"][pos:]

            #     # see if we got a list of outputs encoded as a string
            #     parts = re.split(r"{{!--GMARKERmany[^}]+}}", return_value)
            #     if len(parts) > 1:
            #         return parts[1:-1]
            #     else:
            #         return return_value
            # else:
            #     return ""

        elif node_name == 'block_command_call':
            parts = [await self.visit(child, variable_stack) for child in node.children]
            if len(parts) > 1:
                command_name, args = parts
            else:
                command_name = parts[0]
                args = []
            return command_name, args

        elif node_name == 'block_command_open':
            return await self.visit(node.children[2], variable_stack)
            # visited_children = [await self.visit(child) for child in node.children]
            # return visited_children[2]

        elif node_name == 'block_command':

            # if execution is already stopped before we start the command block we just return unchanged
            if not self.executing:
                variable_stack["@raw_prefix"] += node.text
                return ""

            # create a block content variable
            # block_content = node["block_content"]
            # block_content = [node.children[1]]
            # for child in node.children[2].children:
            #     if child.text == '':
            #         continue
            #     block_content.append(child.children[0])
            #     block_content.append(child.children[1])
            assert node[1].get_name() == "block_content" # TODO: figure out why node["block_content"] doesn't work (has to do with SavedText messing up the keys)
            self.block_content.append(node[1])

            # get the command name and arguments
            call = node["command_call"]
            command_name = call["name"]
            command_args = [await self.visit(arg, variable_stack) for arg in call[1:]]
            
            # command_args = []
            # if "positional_command_arg" in call:
            #     for arg in call["positional_command_arg"]:
            #         command_args.append(await self.visit(arg, variable_stack))
            # if "named_command_arg" in call:
            #     for arg in call["named_command_arg"]:
            #         command_args.append(await self.visit(arg, variable_stack))
            # command_name, command_args = [await self.visit(arg, variable_stack) for arg in node["command_call"]["args"]]

            # make sure we have a matching end command TODO: move this to a parser action
            # if not (node.text.endswith("/"+command_name+"}}") or node.text.endswith("/"+command_name+"~}}")):
            #     raise SyntaxError("Compiler command block starting with `"+node.text[:20]+"...` does not end with a matching `{{/"+command_name+"}}` but instead ends with `..."+node.text[-20:]+"!")

            # if execution stops while parsing the start command just return unchanged
            if not self.executing:
                variable_stack["@raw_prefix"] += node.text
                return ""

            # add the start marker
            escaped_node_text = node.text.replace("$", "&#36;").replace("{", "&#123;").replace("}", "&#125;")
            start_marker = "{{!--"+f"GMARKER_START_{command_name}${escaped_node_text}$"+"--}}"
            variable_stack["@raw_prefix"] += start_marker

            if command_name in variable_stack:
                command_function = variable_stack[command_name]
                positional_args = []
                named_args = {}
                for arg in command_args:
                    if isinstance(arg, PositionalArgument):
                        positional_args.append(arg.value)
                    elif isinstance(arg, NamedArgument):
                        named_args[arg.name] = arg.value
                
                # see if the command expects parser context
                sig = inspect.signature(command_function)
                if "_parser_context" in sig.parameters:
                    named_args["_parser_context"] = {
                        "parser": self,
                        "block_content": self.block_content[-1],
                        "variable_stack": variable_stack,
                        "parser_node": node,
                        "block_close_node": node[-1],
                        "next_node": next_node,
                        "next_next_node": next_next_node,
                        "prev_node": node[0]
                    }
                
                # call the optionally asyncronous command
                if self._logging:
                    self.program.log.append({
                        "type": "start",
                        "name": command_name,
                        "positional_args": positional_args,
                        "named_args": {k:v for k,v in named_args.items() if k != "_parser_context"},
                        "@prefix": variable_stack["@prefix"],
                        # "node_id": id(node)
                    })
                    pos = len(variable_stack["@prefix"])
                if inspect.iscoroutinefunction(command_function):
                    command_output = await command_function(*positional_args, **named_args)
                else:
                    command_output = command_function(*positional_args, **named_args)
                if self._logging:
                    self.program.log.append({"type": "end", "name": command_name, "new_prefix": variable_stack["@prefix"][pos:]})

                # if the command didn't send partial output we do it here
                if command_output is not None:
                    variable_stack["@raw_prefix"] += command_output

            # pop off the block content after the command call
            self.block_content.pop()

            variable_stack["@raw_prefix"] += "{{!--" + f"GMARKER_END_{command_name}$$" + "--}}"
            return

        else:
            visited_children = []
            for i, child in enumerate(node):
                if len(node) > i + 1:
                    inner_next_node = node[i + 1]
                else:
                    inner_next_node = next_node
                
                if len(node) > i + 2:
                    inner_next_next_node = node[i + 2]
                elif len(node) == i + 2:
                    inner_next_next_node = next_node
                else:
                    inner_next_next_node = next_next_node
                
                if i > 0:
                    inner_prev_node = node[i - 1]
                else:
                    inner_prev_node = prev_node
                visited_children.append(await self.visit(child, variable_stack, inner_next_node, inner_next_next_node, inner_prev_node, node, parent_node))
            # visited_children = [self.visit(child) for child in node.children]
            
            if len(visited_children) == 1:
                return visited_children[0]
            else:
                return "".join("" if c is None else c for c in visited_children)

    # def get_variable(self, name, default_value=None):
    #     parts = re.split(r"\.|\[", name) 40 ms 2048 12B
    #     for variables in reversed(self.variable_stack):
    #         curr_pos = variables
    #         found = True
    #         for part in parts:
    #             if part.endswith("]"):
    #                 var_part = ast.literal_eval(part[:-1])
    #             else:
    #                 var_part = part
    #             try:
    #                 next_pos = curr_pos[var_part]
    #                 next_found = True
    #             except KeyError:
    #                 next_found = False
    #             if next_found:
    #                 curr_pos = next_pos
    #             else:
    #                 found = False
    #                 break
    #         if found:
    #             return curr_pos
    #     return default_value # variable not found

    # def variable_exists(self, name):
    #     return self.get_variable(name, _NO_VALUE) != _NO_VALUE

    # def set_variable(self, name, value):
    #     parts = re.split(r"\.|\[", name)
    #     # parts = name.split(".")
    #     found = True
    #     for variables in reversed(self.variable_stack):
    #         curr_pos = variables
    #         found = True
    #         for part in parts:
    #             if part.endswith("]"):
    #                 var_part = ast.literal_eval(part[:-1])
    #             else:
    #                 var_part = part
    #             try:
    #                 next_pos = curr_pos[var_part]
    #                 next_found = True
    #             except KeyError:
    #                 next_found = False
                
    #             if next_found:
    #                 if part == parts[-1]:
    #                     curr_pos[var_part] = value
    #                     break
    #                 else:
    #                     curr_pos = next_pos
    #             else:
    #                 if part == parts[-1] and len(parts) > 1: # setting a new property
    #                     curr_pos[var_part] = value
    #                 else:
    #                     found = False
    #                 break
    #         if found:
    #             break
    #     if not found:
    #         assert len(parts) == 1, "Can't set a property of a non-existing variable: " + name
    #         self.variable_stack[0][name] = value

    # def extend_prefix(self, text, variable_stack):
    #     if text == "" or text is None:
    #         return
    #     variable_stack["@raw_prefix"] += str(text)
    #     self.program.update_display()
    
    # def reset_prefix(self, pos):
    #     self.prefix = self.prefix[:pos]
    #     self.program.update_display()
    #     # TODO: undo the echo if needed

class PositionalArgument:
    def __init__(self, value):
        self.value = value

class NamedArgument:
    def __init__(self, name, value):
        self.name = name
        self.value = value

class StopCompletion(Exception):
    pass

class SyntaxException(Exception):
    def __init__(self, msg, pyparsing_exception=None):
        super().__init__(msg)
        self.pyparsing_exception = pyparsing_exception


from ._openai import OpenAI, MSALOpenAI, AzureOpenAI
from ._transformers import Transformers
from ._mock import Mock
from ._llm import LLM, LLMSession, SyncSession
from ._deep_speed import DeepSpeed
from . import transformers
from dotagent import caches



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

import ctypes
# libgcc_s = ctypes.CDLL('libgcc_s.so.1')

from collections import defaultdict
from concurrent.futures import as_completed, ProcessPoolExecutor
import logging

from dotagent.fact_check.code.helper._execution import test_case_against_solution

logging.basicConfig(
    format="SystemLog: [%(asctime)s][%(name)s][%(levelname)s] - %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
    level=logging.INFO,
)

logger = logging.getLogger(__name__)

def evaluate_test_cases_multi_solution(prompt, testcases_input,
                                       multi_solutions, timeout=0.1):
    logger.info(f'Start evaluation with test code, timeout={timeout}')

    with ProcessPoolExecutor() as executor:
        futures = []
        results = [[None for _ in multi_solutions] for _ in testcases_input]

        for i, testcase in enumerate(testcases_input):
            for j, solution in enumerate(multi_solutions):
                args = (prompt, solution, testcase, timeout)
                future = executor.submit(test_case_against_solution, *args)
                futures.append((i, j, future))
        logger.info(f'{len(futures)} execution requests are submitted')
        
        for completed_future in as_completed([f[2] for f in futures]):
            for i, j, future in futures:
                if future == completed_future:
                    logger.info('[{}/{}] execution completed'.format(
                        i * len(multi_solutions) + j + 1, len(futures)))
                    result = completed_future.result()
                    results[i][j] = result
                    break

    return results









import asyncio
import os
import types
import httpx

from dotagent.logger import get_logger
from dotagent.utils import Singleton
from dotagent.audio.text_to_speech.base import TextToSpeech

logger = get_logger(__name__)
DEBUG = False

config = types.SimpleNamespace(**{
    'chunk_size': 1024,
    'url': 'https://api.elevenlabs.io/v1/text-to-speech/{voice_id}/stream',
    'headers': {
        'Accept': 'audio/mpeg',
        'Content-Type': 'application/json',
        'xi-api-key': os.environ['ELEVEN_LABS_API_KEY']
    },
    'data': {
        'model_id': 'eleven_monolingual_v1',
        'voice_settings': {
            'stability': 0.5,
            'similarity_boost': 0.75
        }
    }
})


class ElevenLabs(Singleton, TextToSpeech):
    def __init__(self):
        super().__init__()
        logger.info("Initializing [ElevenLabs Text To Speech] voices...")

    async def stream(self, text, websocket, tts_event: asyncio.Event, voice_id="21m00Tcm4TlvDq8ikWAM",
                     first_sentence=False, language='en-US') -> None:
        if DEBUG:
            return
        if voice_id == "":
            logger.info(
                f"voice_id is not found in .env file, using ElevenLabs default voice")
            voice_id = "21m00Tcm4TlvDq8ikWAM"
        headers = config.headers
        if language != 'en-US':
            config.data["model_id"] = 'eleven_multilingual_v1'
        data = {
            "text": text,
            **config.data,
        }
        url = config.url.format(voice_id=voice_id)
        if first_sentence:
            url = url + '?optimize_streaming_latency=4'
        async with httpx.AsyncClient() as client:
            response = await client.post(url, json=data, headers=headers)
            if response.status_code != 200:
                logger.error(
                    f"ElevenLabs returns response {response.status_code}")
            async for chunk in response.aiter_bytes():
                await asyncio.sleep(0.1)
                if tts_event.is_set():
                    # stop streaming audio
                    break
                await websocket.send_bytes(chunk)



"""Wrapper around Activeloop Deep Lake."""
from __future__ import annotations

import logging
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, Union

import numpy as np

try:
    import deeplake
    from deeplake.core.fast_forwarding import version_compare
    from deeplake.core.vectorstore import DeepLakeVectorStore

    _DEEPLAKE_INSTALLED = True
except ImportError:
    _DEEPLAKE_INSTALLED = False

from dotagent.schema import Document
from dotagent.vectorstores.embeddings.base import Embeddings
from dotagent.vectorstores.base import VectorStore
# from dotagent.vectorstores.utils import maximal_marginal_relevance

logger = logging.getLogger(__name__)

DEFAULT_DATASET_PATH = "./deeplake/"

class DeepLake(VectorStore):
    """Wrapper around Deep Lake, a data lake for deep learning applications.

    To use, you should have the ``deeplake`` python package installed.

    Example:
        .. code-block:: python

                from dotagent.vectorstores import DeepLake
                from dotagent.embeddings.openai import OpenAIEmbeddings

                embeddings = OpenAIEmbeddings()
                vectorstore = DeepLake("dotagent_db_path", embeddings.embed_query)
    """

    def __init__(
        self,
        dataset_path: str = DEFAULT_DATASET_PATH,
        token: Optional[str] = None,
        embedding_function: Optional[Embeddings] = None,
        read_only: bool = False,
        ingestion_batch_size: int = 1000,
        num_workers: int = 0,
        verbose: bool = True,
        exec_option: str = "python",
        **kwargs: Any,
    ) -> None:
        """Creates an empty DeepLakeVectorStore or loads an existing one.

        The DeepLakeVectorStore is located at the specified ``path``.

        Examples:
            >>> # Create a vector store with default tensors
            >>> deeplake_vectorstore = DeepLake(
            ...        path = <path_for_storing_Data>,
            ... )
            >>>
            >>> # Create a vector store in the Deep Lake Managed Tensor Database
            >>> data = DeepLake(
            ...        path = "hub://org_id/dataset_name",
            ...        exec_option = "tensor_db",
            ... )

        Args:
            dataset_path (str): Path to existing dataset or where to create
                a new one. Defaults to DEFAULT_DATASET_PATH
            token (str, optional):  Activeloop token, for fetching credentials
                to the dataset at path if it is a Deep Lake dataset.
                Tokens are normally autogenerated. Optional.
            embedding_function (str, optional): Function to convert
                either documents or query. Optional.
            read_only (bool): Open dataset in read-only mode. Default is False.
            ingestion_batch_size (int): During data ingestion, data is divided
                into batches. Batch size is the size of each batch.
                Default is 1000.
            num_workers (int): Number of workers to use during data ingestion.
                Default is 0.
            verbose (bool): Print dataset summary after each operation.
                Default is True.
            exec_option (str): DeepLakeVectorStore supports 3 ways to perform
                searching - "python", "compute_engine", "tensor_db".
                Default is "python".
                - ``python`` - Pure-python implementation that runs on the client.
                WARNING: using this with big datasets can lead to memory
                issues. Data can be stored anywhere.
                - ``compute_engine`` - C++ implementation of the Deep Lake Compute
                Engine that runs on the client. Can be used for any data stored in
                or connected to Deep Lake. Not for in-memory or local datasets.
                - ``tensor_db`` - Hosted Managed Tensor Database that is
                responsible for storage and query execution. Only for data stored in
                the Deep Lake Managed Database. Use runtime = {"db_engine": True} during
                dataset creation.
            **kwargs: Other optional keyword arguments.

        Raises:
            ValueError: If some condition is not met.
        """

        self.ingestion_batch_size = ingestion_batch_size
        self.num_workers = num_workers
        self.verbose = verbose

        if _DEEPLAKE_INSTALLED is False:
            raise ValueError(
                "Could not import deeplake python package. "
                "Please install it with `pip install deeplake`."
            )

        if version_compare(deeplake.__version__, "3.6.2") == -1:
            raise ValueError(
                "deeplake version should be >= 3.6.3, but you've installed"
                f" {deeplake.__version__}. Consider upgrading deeplake version \
                    pip install --upgrade deeplake."
            )
        self.dataset_path = dataset_path

        self.vectorstore = DeepLakeVectorStore(
            path=self.dataset_path,
            embedding_function=embedding_function,
            read_only=read_only,
            token=token,
            exec_option=exec_option,
            verbose=verbose,
            **kwargs,
        )

        self._embedding_function = embedding_function
        self._id_tensor_name = "ids" if "ids" in self.vectorstore.tensors() else "id"

    def add_texts(
        self,
        texts: Iterable[str],
        metadatas: Optional[List[dict]] = None,
        ids: Optional[List[str]] = None,
        **kwargs: Any,
    ) -> List[str]:
        """Run more texts through the embeddings and add to the vectorstore.

        Examples:
            >>> ids = deeplake_vectorstore.add_texts(
            ...     texts = <list_of_texts>,
            ...     metadatas = <list_of_metadata_jsons>,
            ...     ids = <list_of_ids>,
            ... )

        Args:
            texts (Iterable[str]): Texts to add to the vectorstore.
            metadatas (Optional[List[dict]], optional): Optional list of metadatas.
            ids (Optional[List[str]], optional): Optional list of IDs.
            **kwargs: other optional keyword arguments.

        Returns:
            List[str]: List of IDs of the added texts.
        """
        kwargs = {}
        if ids:
            if self._id_tensor_name == "ids":  # for backwards compatibility
                kwargs["ids"] = ids
            else:
                kwargs["id"] = ids

        if metadatas is None:
            metadatas = [{}] * len(list(texts))

        return self.vectorstore.add(
            text=texts,
            metadata=metadatas,
            embedding_data=texts,
            embedding_tensor="embedding",
            embedding_function=kwargs.get("embedding_function")
            or self._embedding_function.embed_documents,  # type: ignore
            return_ids=True,
            **kwargs,
        )

    def _search_tql(
        self,
        tql_query: Optional[str],
        exec_option: Optional[str] = None,
        return_score: bool = False,
    ) -> Any[List[Document], List[Tuple[Document, float]]]:
        """Function for performing tql_search.

        Args:
            tql_query (str): TQL Query string for direct evaluation.
                Available only for `compute_engine` and `tensor_db`.
            exec_option (str, optional): Supports 3 ways to search.
                Could be "python", "compute_engine" or "tensor_db". Default is "python".
                - ``python`` - Pure-python implementation for the client.
                    WARNING: not recommended for big datasets due to potential memory
                    issues.
                - ``compute_engine`` - C++ implementation of Deep Lake Compute
                    Engine for the client. Not for in-memory or local datasets.
                - ``tensor_db`` - Hosted Managed Tensor Database for storage
                    and query execution. Only for data in Deep Lake Managed Database.
                        Use runtime = {"db_engine": True} during dataset creation.
            return_score (bool): Return score with document. Default is False.

        Returns:
            List[Document] - A list of documents

        Raises:
            ValueError: If return_score is True but some condition is not met.
        """
        result = self.vectorstore.search(
            query=tql_query,
            exec_option=exec_option,
        )
        metadatas = result["metadata"]
        texts = result["text"]

        docs = [
            Document(
                page_content=text,
                metadata=metadata,
            )
            for text, metadata in zip(texts, metadatas)
        ]

        if return_score:
            raise ValueError("scores can't be returned with tql search")

        return docs

    def _search(
        self,
        query: Optional[str] = None,
        embedding: Optional[Union[List[float], np.ndarray]] = None,
        embedding_function: Optional[Callable] = None,
        k: int = 4,
        distance_metric: str = "L2",
        filter: Optional[Union[Dict, Callable]] = None,
        return_score: bool = True,
        exec_option: Optional[str] = None,
        **kwargs: Any,
    ) -> Any[List[Document], List[Tuple[Document, float]]]:
        """
        Return docs similar to query.

        Args:
            query (str, optional): Text to look up similar docs.
            embedding (Union[List[float], np.ndarray], optional): Query's embedding.
            embedding_function (Callable, optional): Function to convert `query`
                into embedding.
            k (int): Number of Documents to return.
            distance_metric (str): `L2` for Euclidean, `L1` for Nuclear, `max`
                for L-infinity distance, `cos` for cosine similarity, 'dot' for dot
                product.
            filter (Union[Dict, Callable], optional): Additional filter prior
                to the embedding search.
                - ``Dict`` - Key-value search on tensors of htype json, on an
                    AND basis (a sample must satisfy all key-value filters to be True)
                    Dict = {"tensor_name_1": {"key": value},
                            "tensor_name_2": {"key": value}}
                - ``Function`` - Any function compatible with `deeplake.filter`.
            use_maximal_marginal_relevance (bool): Use maximal marginal relevance.
            fetch_k (int): Number of Documents for MMR algorithm.
            return_score (bool): Return the score.
            exec_option (str, optional): Supports 3 ways to perform searching.
                Could be "python", "compute_engine" or "tensor_db".
                - ``python`` - Pure-python implementation for the client.
                    WARNING: not recommended for big datasets.
                - ``compute_engine`` - C++ implementation of Deep Lake Compute
                    Engine for the client. Not for in-memory or local datasets.
                - ``tensor_db`` - Hosted Managed Tensor Database for storage
                    and query execution. Only for data in Deep Lake Managed Database.
                    Use runtime = {"db_engine": True} during dataset creation.
            **kwargs: Additional keyword arguments.

        Returns:
            List of Documents by the specified distance metric,
            if return_score True, return a tuple of (Document, score)

        Raises:
            ValueError: if both `embedding` and `embedding_function` are not specified.
        """

        if kwargs.get("tql_query"):
            return self._search_tql(
                tql_query=kwargs["tql_query"],
                exec_option=exec_option,
                return_score=return_score,
            )

        if embedding_function:
            if isinstance(embedding_function, Embeddings):
                _embedding_function = embedding_function.embed_query
            else:
                _embedding_function = embedding_function
        elif self._embedding_function:
            _embedding_function = self._embedding_function.embed_query
        else:
            _embedding_function = None

        if embedding is None:
            if _embedding_function is None:
                raise ValueError(
                    "Either `embedding` or `embedding_function` needs to be specified."
                )

            embedding = _embedding_function(query) if query else None

        if isinstance(embedding, list):
            embedding = np.array(embedding, dtype=np.float32)
            if len(embedding.shape) > 1:
                embedding = embedding[0]

        result = self.vectorstore.search(
            embedding=embedding,
            k=k,
            distance_metric=distance_metric,
            filter=filter,
            exec_option=exec_option,
            return_tensors=["embedding", "metadata", "text"],
        )

        scores = result["score"]
        embeddings = result["embedding"]
        metadatas = result["metadata"]
        texts = result["text"]


        for meta, embed in zip(metadatas, embeddings):
            meta['embedding'] = embed
            metadatas.append(meta)

        docs = [
            Document(
                page_content=text,
                metadata=metadata,
            )
            for text, metadata in zip(texts, metadatas)
        ]

        if return_score:
            return [(doc, score) for doc, score in zip(docs, scores)]

        return docs

    def similarity_search(
        self,
        query: Optional[str] = None,
        embedding: Optional[List[float]] = None,
        k: int = 4,
        **kwargs: Any,
    ) -> List[Document]:
        """
        Return docs most similar to query.

        Examples:
            >>> # Search using an embedding
            >>> data = vector_store.similarity_search(
            ...     query=<your_query>,
            ...     k=<num_items>,
            ...     exec_option=<preferred_exec_option>,
            ... )
            >>> # Run tql search:
            >>> data = vector_store.tql_search(
            ...     tql_query="SELECT * WHERE id == <id>",
            ...     exec_option="compute_engine",
            ... )
        """
        if (embedding is None and query is None) or (embedding is not None and query is not None):
            raise ValueError("You must provide either query embeddings or query texts, but not both")


        return self._search(
            query=query,
            embedding=embedding,
            k=k,
            **kwargs,
        )


    @classmethod
    def from_texts(
        cls,
        texts: List[str],
        embedding_function: Optional[Embeddings] = None,
        metadatas: Optional[List[dict]] = None,
        ids: Optional[List[str]] = None,
        dataset_path: str = DEFAULT_DATASET_PATH,
        **kwargs: Any,
    ) -> DeepLake:
        """Create a Deep Lake dataset from a raw documents.

        If a dataset_path is specified, the dataset will be persisted in that location,
        otherwise by default at `./deeplake`

        Examples:
        >>> # Search using an embedding
        >>> vector_store = DeepLake.from_texts(
        ...        texts = <the_texts_that_you_want_to_embed>,
        ...        embedding_function = <embedding_function_for_query>,
        ...        k = <number_of_items_to_return>,
        ...        exec_option = <preferred_exec_option>,
        ... )

        """
        
        if kwargs.get("embedding"):
            raise ValueError(
                "using embedding as embedidng_functions is deprecated. "
                "Please use `embedding_function` instead."
            )

        deeplake_dataset = cls(
            dataset_path=dataset_path, embedding_function=embedding_function, **kwargs
        )
        deeplake_dataset.add_texts(
            texts=texts,
            metadatas=metadatas,
            ids=ids,
            embedding_function=embedding.embed_documents,  # type: ignore
        )
        return deeplake_dataset

    def delete(
        self,
        ids: Any[List[str], None] = None,
        filter: Any[Dict[str, str], None] = None,
        delete_all: Any[bool, None] = None,
    ) -> bool:
        """Delete the entities in the dataset.

        Args:
            ids (Optional[List[str]], optional): The document_ids to delete.
                Defaults to None.
            filter (Optional[Dict[str, str]], optional): The filter to delete by.
                Defaults to None.
            delete_all (Optional[bool], optional): Whether to drop the dataset.
                Defaults to None.

        Returns:
            bool: Whether the delete operation was successful.
        """
        self.vectorstore.delete(
            ids=ids,
            filter=filter,
            delete_all=delete_all,
        )

        return True

    @classmethod
    def force_delete_by_path(cls, path: str) -> None:
        """Force delete dataset by path.

        Args:
            path (str): path of the dataset to delete.

        Raises:
            ValueError: if deeplake is not installed.
        """

        try:
            import deeplake
        except ImportError:
            raise ValueError(
                "Could not import deeplake python package. "
                "Please install it with `pip install deeplake`."
            )
        deeplake.delete(path, large_ok=True, force=True)

    def delete_dataset(self) -> None:
        """Delete the collection."""
        self.delete(delete_all=True)



from typing import Any, Dict, List
from abc import ABC, abstractmethod
from dotagent.schema import BaseMessage
from pydantic import BaseModel, Field


class BaseMemory(ABC, BaseModel):
    """Abstract base class for memory management in a chatbot application. 
    This class defines the interface for setting, getting, and checking existence of data in memory.
    """

    # All memories are stored in this list
    messages: List[Dict[BaseMessage, Any]] = Field(default=list())

    @abstractmethod
    def add_memory(self, prompt: str, llm_response: Any) -> None:
        """Add a memory to the store."""
        pass

    @abstractmethod
    def get_memory(self, **kwargs) -> str:
        """Retrieve entire memory from the store."""
        pass

    @abstractmethod
    def remove_memory(self, prompt: str) -> None:
        """Remove a memory from the store."""
        pass

    @abstractmethod
    def clear(self) -> None:
        """Clear all memories."""
        pass

    @property
    def memory_prompts(self) -> List[str]:
        """Return the prompts for all memories."""
        prompts = []
        for conversation in self.messages:
            prompts.append(conversation['prompt'])
        
        return prompts


from dotagent.tools.toolkits.requests_toolkit.requests.base import RequestsGetTool, TextRequestsWrapper

# Instantiate the RequestsGetTool with TextRequestsWrapper as requests_wrapper
get_tool = RequestsGetTool(requests_wrapper=TextRequestsWrapper())

# The URL you want to get data from
url = "https://yaml.org"

# Use the get method
response = get_tool._run(url)

# Output the response
print(response)


import os
import openai

from ._base import BaseImageLlm


class OpenAiDalle(BaseImageLlm):
    def __init__(self, api_key, image_model=None, number_of_results=1):
        """
        Args:
            api_key (str): The OpenAI API key.
            image_model (str): The image model.
            number_of_results (int): The number of results.
        """
        self.number_of_results = number_of_results
        self.api_key = api_key
        self.image_model = image_model
        openai.api_key = api_key
        openai.api_base = os.getenv("OPENAI_API_BASE", "https://api.openai.com/v1")

    def get_image_model(self):
        """
        Returns:
            str: The image model.
        """
        return self.image_model

    def generate_image(self, prompt: str, size: int = 512):
        """
        Call the OpenAI image API.

        Args:
            prompt (str): The prompt.
            size (int): The size.
            num (int): The number of images.

        Returns:
            dict: The response.
        """
        response = openai.Image.create(
            prompt=prompt,
            n=self.number_of_results,
            size=f"{size}x{size}"
        )
        return response



"""Tools for making requests to an API endpoint."""

from pydantic import BaseModel
from typing import List

from dotagent.tools.toolkits.requests_toolkit.requests.utils import TextRequestsWrapper
from dotagent.tools.basetool import BaseTool
from dotagent.tools.toolkits.base import BaseToolkit
from dotagent.tools.toolkits.requests_toolkit.requests.base import RequestsGetTool, RequestsPostTool, RequestsPatchTool, RequestsPutTool, RequestsDeleteTool


class RequestsToolkit(BaseToolkit):
    """Base class for requests tools."""

    requests_wrapper: TextRequestsWrapper

    def get_tools(self) -> List[BaseTool]:
        """Get the tools in the toolkit."""
        return [
            RequestsGetTool(requests_wrapper=self.requests_wrapper),
            RequestsPostTool(requests_wrapper=self.requests_wrapper),
            RequestsPatchTool(requests_wrapper=self.requests_wrapper),
            RequestsPutTool(requests_wrapper=self.requests_wrapper),
            RequestsDeleteTool(requests_wrapper=self.requests_wrapper),
        ]


import sys
import io
import yaml
import pdb

class python_executor:
    def __init__(self):
        pass

    def run_single(self, program):
        buffer = io.StringIO() # Create an in-memory buffer for the output
        stdout = sys.stdout # Save the original standard output
        sys.stdout = buffer # Redirect the standard output to the buffer
        try:
            exec(program)
        except Exception as e:
            # Handle the error here
            error_message = str(e)
            sys.stdout = stdout  # Restore the original standard output
            return error_message

        exec(program)
        sys.stdout = stdout # Restore the original standard output
        output = buffer.getvalue() # Get the output from the buffer
        return output

    def run(self, snippet):
        if snippet == None:
            return None
        exec_result = self.run_single(snippet)
        if exec_result and 'False' in exec_result:
            exec_result = 'False'
        else:
            exec_result = 'True'
        return exec_result


"""Chroma Reader."""

from typing import Any

from dotagent.knowledgebase.document_loaders.basereader import BaseReader
from dotagent.schema import DocumentNode

class ChromaReader(BaseReader):
    """Chroma reader.

    Retrieve documents from existing persisted Chroma collections.

    Args:
        collection_name: Name of the peristed collection.
        persist_directory: Directory where the collection is persisted.

    """

    def __init__(
        self,
        collection_name: str,
        persist_directory: str,
    ) -> None:
        """Initialize with parameters."""
        import chromadb  # noqa: F401
        from chromadb.config import Settings

        self.collection_name = collection_name

        if (collection_name is None) or (persist_directory is None):
            raise ValueError("Please provide a collection name and persist directory.")

        self._client = chromadb.Client(
            Settings(
                is_persistent=True, persist_directory=persist_directory
            )
        )
        self._collection = self._client.get_collection(collection_name)

    def load_data(
        self,
        query_vector: Any,
        limit: int = 10,
    ) -> Any:
        """Load data from Chroma.

        Args:
            query_vector (Any): Query
            limit (int): Number of results to return.

        Returns:
            List[DocumentNode]: A list of documents.
        """
        results = self._collection.query(query_embeddings=query_vector, n_results=limit)

        metadata = {
            "collection_name": self.collection_name,
            "query_vector": query_vector,
            "limit": limit,
        }
        documents = []
        for result in zip(results["ids"], results["documents"], results["embeddings"]):
            doc = DocumentNode(
                doc_id=result[0][0],
                text=result[1][0],
                embedding=result[2][0],
                extra_info=metadata  
            )
            documents.append(doc)

        return documents



from __future__ import annotations

"""Base implementation for tools or skills."""

import warnings
from abc import ABC, abstractmethod
from inspect import signature
from typing import Any, Callable, Dict, Optional, Tuple, Type, Union

from pydantic import (
    BaseModel,
    create_model,
    root_validator,
    validate_arguments,
)


def _get_filtered_args(
    inferred_model: Type[BaseModel],
    func: Callable,
) -> dict:
    """Get the arguments from a function's signature."""
    schema = inferred_model.schema()["properties"]
    valid_keys = signature(func).parameters
    return {k: schema[k] for k in valid_keys if k != "run_manager"}

def _create_subset_model(
    name: str, model: BaseModel, field_names: list
) -> Type[BaseModel]:
    """Create a pydantic model with only a subset of model's fields."""
    fields = {}
    for field_name in field_names:
        field = model.__fields__[field_name]
        fields[field_name] = (field.type_, field.field_info)
    return create_model(name, **fields)  # type: ignore

def create_schema_from_function(
    model_name: str,
    func: Callable,
) -> Type[BaseModel]:
    """
        model_name: Name to assign to the generated pydandic schema
        func: Function to generate the schema from
    """
    validated = validate_arguments(func, config=_SchemaConfig)  # type: ignore
    inferred_model = validated.model  # type: ignore
    if "run_manager" in inferred_model.__fields__:
        del inferred_model.__fields__["run_manager"]
    # Pydantic adds placeholder virtual fields we need to strip
    valid_properties = _get_filtered_args(inferred_model, func)
    return _create_subset_model(
        f"{model_name}Schema", inferred_model, list(valid_properties)
    )


class BaseTool(ABC, BaseModel):
  # Interface dotagent tools must implement.

    name: str
    #The unique name of the tool that clearly communicates its purpose.
    description: str
    #Used to tell the model how/when/why to use the tool.You can provide few-shot examples as a part of the description.
    func: Callable = None
    #Function which acts as a tool and takes in input
    args_schema: Optional[Type[BaseModel]] = None
    #Pydantic model class to validate and parse the tool's input arguments
    return_direct: bool = False
    # Whether to return the tool's output directly. Setting this to True means that after the tool is called, the AgentExecutor will stop looping.
    verbose: bool = False
    #Whether to log the tool's progress.

    @property
    def is_single_input(self) -> bool:
        """Whether the tool only accepts a single input."""
        keys = {k for k in self.args if k != "kwargs"}
        return len(keys) == 1

    @property
    def args(self) -> dict:
        if self.args_schema is not None:
            return self.args_schema.schema()["properties"]
        else:
            schema = create_schema_from_function(self.name, self.run)
            return schema.schema()["properties"]

    def _parse_input(
        self,
        tool_input: Union[str, Dict],
    ) -> Union[str, Dict[str, Any]]:
        input_args = self.args_schema
        if isinstance(tool_input, str):
            if input_args is not None:
                key_ = next(iter(input_args.__fields__.keys()))
                input_args.validate({key_: tool_input})
            return tool_input
        else:
            if input_args is not None:
                result = input_args.parse_obj(tool_input)
                return {k: v for k, v in result.dict().items() if k in tool_input}
        return tool_input

    @root_validator()
    def raise_deprecation(cls, values: Dict) -> Dict:
        """Raise deprecation warning if callback_manager is used."""
        if values.get("callback_manager") is not None:
            warnings.warn(
                "callback_manager is deprecated. Please use callbacks instead.",
                DeprecationWarning,
            )
            values["callbacks"] = values.pop("callback_manager", None)
        return values

    @abstractmethod
    def run(
        self,
        *args: Any,
        **kwargs: Any,
    ) -> Any:
        """Use the tool.
        """

    def _to_args_and_kwargs(self, tool_input: Union[str, Dict]) -> Tuple[Tuple, Dict]:
        if isinstance(tool_input, str):
            return (tool_input,), {}
        else:
            return (), tool_input

    def run(self,tool_input: Union[str, Dict], verbose: Optional[bool] = None,**kwargs: Any) -> Any:
        """Parses the output and checks if the input is string and handles exceptions"""
        parsed_input = self._parse_input(tool_input)

        observation = self.run(parsed_input)

        return observation

        # if isinstance(parsed_input, str):
        #     raise Exception("Tool input should be string")
        # else:
        #     return parsed_input

class ExceptionTool(BaseTool):
    name = "_Exception"
    description = "Exception tool"

    def run(
        self,
        query: str,
    ) -> str:
        return query
    
class InvalidTool(BaseTool):
    """Tool that is run when invalid tool name is encountered by agent."""

    name = "invalid_tool"
    description = "Called when tool name is invalid."

    def run(
        self, tool_name: str,
    ) -> str:
        """Use the tool."""
        return f"{tool_name} is not a valid tool, try another one."

class Tool(BaseTool):
    """Tool that takes in function or coroutine directly."""

    class Config:
        arbitrary_types_allowed = True

    def args(self) -> dict:
        """The tool's input arguments."""
        if self.args_schema is not None:
            return self.args_schema.schema()["properties"]
        # For backwards compatibility, if the function signature is ambiguous,
        # assume it takes a single string input.
        return {"tool_input": {"type": "string"}}
    
    def run(self, tool_input, **kwargs: Any) -> Any:
        """Actually calls the tool and gives output."""
        try:
            return (self.func(tool_input, **kwargs))
        except Exception as e:
            return e

    def __call__(self, tool_input: str, **kwargs) -> str:
        """Make tool callable."""
        # try:
        #     parsed_input = self.run(tool_input , **kwargs)
        # except Exception as e:
        #     return e
        # final_result =  self.run(tool_input=parsed_input, **kwargs)
        # return final_result


from dotagent.knowledgebase.text_retrievers.arxiv import ArxivRetriever
from dotagent.knowledgebase.text_retrievers.aws_kendra import AwsKendraIndexRetriever
from dotagent.knowledgebase.text_retrievers.azure_cognitive import AzureCognitiveSearchRetriever
from dotagent.knowledgebase.text_retrievers.chatgpt_plugin import ChatGPTPluginRetriever
from dotagent.knowledgebase.text_retrievers.contextual_compression import ContextualCompressionRetriever
from dotagent.knowledgebase.text_retrievers.databerry import DataberryRetriever
from dotagent.knowledgebase.text_retrievers.elastic_search import ElasticSearchBM25Retriever
from dotagent.knowledgebase.text_retrievers.knn import KNNRetriever
from dotagent.knowledgebase.text_retrievers.llama_index import (
    LlamaIndexGraphRetriever,
    LlamaIndexRetriever,
)
from dotagent.knowledgebase.text_retrievers.merger import MergerRetriever
from dotagent.knowledgebase.text_retrievers.metal import MetalRetriever
from dotagent.knowledgebase.text_retrievers.pinecone import PineconeHybridSearchRetriever
from dotagent.knowledgebase.text_retrievers.pupmed import PubMedRetriever
from dotagent.knowledgebase.text_retrievers.remote_retriever import RemotedotagentRetriever
from dotagent.knowledgebase.text_retrievers.svm import SVMRetriever
from dotagent.knowledgebase.text_retrievers.tfidf import TFIDFRetriever
from dotagent.knowledgebase.text_retrievers.time_retriever import TimeWeightedVectorStoreRetriever
from dotagent.knowledgebase.text_retrievers.vespa import VespaRetriever
from dotagent.knowledgebase.text_retrievers.weaviate_hybrid import WeaviateHybridSearchRetriever
from dotagent.knowledgebase.text_retrievers.wikipedia import WikipediaRetriever
from dotagent.knowledgebase.text_retrievers.zep import ZepRetriever

__all__ = [
    "ArxivRetriever",
    "AwsKendraIndexRetriever",
    "AzureCognitiveSearchRetriever",
    "ChatGPTPluginRetriever",
    "ContextualCompressionRetriever",
    "DataberryRetriever",
    "ElasticSearchBM25Retriever",
    "KNNRetriever",
    "LlamaIndexGraphRetriever",
    "LlamaIndexRetriever",
    "MergerRetriever",
    "MetalRetriever",
    "MilvusRetriever",
    "PineconeHybridSearchRetriever",
    "PubMedRetriever",
    "RemotedotagentRetriever",
    "SVMRetriever",
    "TFIDFRetriever",
    "TimeWeightedVectorStoreRetriever",
    "VespaRetriever",
    "WeaviateHybridSearchRetriever",
    "WikipediaRetriever",
    "ZepRetriever",
]


import os

from dotagent.audio.text_to_speech.base import TextToSpeech


def get_text_to_speech(tts: str = None) -> TextToSpeech:
    if not tts:
        tts = os.getenv('TEXT_TO_SPEECH_USE', 'ELEVEN_LABS')
    if tts == 'ELEVEN_LABS':
        from dotagent.audio.text_to_speech.elevenlabs import ElevenLabs
        ElevenLabs.initialize()
        return ElevenLabs.get_instance()
    elif tts == 'GOOGLE_TTS':
        from dotagent.audio.text_to_speech.google_cloud_tts import GoogleCloudTTS
        GoogleCloudTTS.initialize()
        return GoogleCloudTTS.get_instance()
    elif tts == 'UNREAL_SPEECH':
        from dotagent.audio.text_to_speech.unreal_speech import UnrealSpeech
        UnrealSpeech.initialize()
        return UnrealSpeech.get_instance()
    else:
        raise NotImplementedError(f'Unknown text to speech engine: {tts}')



"""Interface for vector stores."""
from __future__ import annotations

import warnings
from abc import ABC, abstractmethod
from typing import (
    Any,
    Iterable,
    List,
    Optional,
    Tuple,
)

from dotagent.schema import Document
from dotagent.vectorstores.embeddings.base import Embeddings


class VectorStore(ABC):
    """Interface for vector stores."""

    @abstractmethod
    def add_texts(
        self,
        texts: Iterable[str],
        metadatas: Optional[List[dict]] = None,
        **kwargs: Any,
    ) -> List[str]:
        """Run more texts through the Embedding and add to the vectorstore.

        Args:
            texts: Iterable of strings to add to the vectorstore.
            metadatas: Optional list of metadatas associated with the texts.
            kwargs: vectorstore specific parameters

        Returns:
            List of ids from adding the texts into the vectorstore.
        """


    def add_documents(self, documents: List[Document], **kwargs: Any) -> List[str]:
        """Run more documents through the Embedding and add to the vectorstore.

        Args:
            documents (List[Document]: Documents to add to the vectorstore.

        Returns:
            List[str]: List of IDs of the added texts.
        """
        # TODO: Handle the case where the user doesn't provide ids on the Collection
        texts = [doc.page_content for doc in documents]
        metadatas = [doc.metadata for doc in documents]
        return self.add_texts(texts, metadatas, **kwargs)



    @abstractmethod
    def similarity_search(
        self, query: str, k: int = 4, **kwargs: Any
    ) -> List[Document]:
        """Return docs most similar to query."""

    def get_matching_text_with_score(
        self,
        query: str,
        k: int = 4,
        **kwargs: Any,
    ) -> List[Tuple[Document, float]]:
        """Return docs and relevance scores in the range [0, 1].

        0 is dissimilar, 1 is most similar.

        Args:
            query: input text
            k: Number of Documents to return. Defaults to 4.
            **kwargs: kwargs to be passed to similarity search. Should include:
                score_threshold: Optional, a floating point value between 0 to 1 to
                    filter the resulting set of retrieved docs

        Returns:
            List of Tuples of (doc, similarity_score)
        """
        docs_and_similarities = self._similarity_search_with_relevance_scores(
            query, k=k, **kwargs
        )
        if any(
            similarity < 0.0 or similarity > 1.0
            for _, similarity in docs_and_similarities
        ):
            warnings.warn(
                "Relevance scores must be between"
                f" 0 and 1, got {docs_and_similarities}"
            )

        score_threshold = kwargs.get("score_threshold")
        if score_threshold is not None:
            docs_and_similarities = [
                (doc, similarity)
                for doc, similarity in docs_and_similarities
                if similarity >= score_threshold
            ]
            if len(docs_and_similarities) == 0:
                warnings.warn(
                    f"No relevant docs were retrieved using the relevance score\
                          threshold {score_threshold}"
                )
        return docs_and_similarities
    

    @classmethod
    def from_documents(
        cls,
        documents: List[Document],
        embedding: Embeddings,
        **kwargs: Any,
    ):
        """Return VectorStore initialized from documents and Embedding."""
        texts = [d.page_content for d in documents]
        metadatas = [d.metadata for d in documents]
        return cls.from_texts(texts, embedding, metadatas=metadatas, **kwargs)


    @classmethod
    @abstractmethod
    def from_texts(
        cls,
        texts: List[str],
        embedding: Embeddings,
        metadatas: Optional[List[dict]] = None,
        **kwargs: Any,
    ):
        """Return VectorStore initialized from texts and Embedding."""



import asyncio
import os
import types
import httpx
import requests
import time

from dotagent.logger import get_logger
from dotagent.utils import Singleton
from dotagent.audio.text_to_speech.base import TextToSpeech

logger = get_logger(__name__)
DEBUG = False

config = types.SimpleNamespace(**{
    'chunk_size': 1024,
    'url': 'https://lab.api.unrealspeech.com/stream',
    'headers': {
        'Accept': 'audio/mpeg',
        'Content-Type': 'application/json',
    },
    'data': {
        'speed': -0.2,
    }
})


class UnrealSpeech(Singleton, TextToSpeech):
    def __init__(self):
        super().__init__()
        logger.info("Initializing [Unreal Speech] voices...")

    async def stream(self, text, websocket, tts_event: asyncio.Event, voice_id=5, *args, **kwargs) -> None:
        if DEBUG:
            return
        params = {
            "text": text,
            "speaker_index": voice_id,
            **config.data,
        }

        async with httpx.AsyncClient() as client:
            response = await client.get(config.url, params=params)
            if response.status_code != 200:
                logger.error(
                    f"Unreal Speech returns response {response.status_code}")
            async for chunk in response.aiter_bytes():
                await asyncio.sleep(0.1)
                if tts_event.is_set():
                    # stop streaming audio
                    break
                await websocket.send_bytes(chunk)



import io
import os
import types
import wave

import speech_recognition as sr
from faster_whisper import WhisperModel
from pydub import AudioSegment
from torch.cuda import is_available as is_cuda_available

from dotagent.audio.speech_to_text.base import SpeechToText
from dotagent.logger import get_logger
from dotagent.utils import Singleton

DEBUG = False
logger = get_logger(__name__)
config = types.SimpleNamespace(**{
    'model': os.getenv("LOCAL_WHISPER_MODEL", "base"),
    'language': 'en',
    'api_key': os.getenv("OPENAI_API_KEY"),
})

# Whisper use a shorter version for language code. Provide a mapping to convert
# from the standard language code to the whisper language code.
WHISPER_LANGUAGE_CODE_MAPPING = {
    "en-US": "en",
    "es-ES": "es",
    "fr-FR": "fr",
    "de-DE": "de",
    "it-IT": "it",
    "pt-PT": "pt",
    "hi-IN": "hi",
    "pl-PL": "pl",
}


class Whisper(Singleton, SpeechToText):
    def __init__(self, use="local"):
        super().__init__()
        if use == "local":
            device = 'cuda' if is_cuda_available() else 'cpu'
            logger.info(f"Loading [Local Whisper] model: [{config.model}]({device}) ...")
            self.model = WhisperModel(
                model_size_or_path=config.model,
                device="auto",
                download_root=None,
            )
        self.recognizer = sr.Recognizer()
        self.use = use
        if DEBUG:
            self.wf = wave.open("output.wav", "wb")
            self.wf.setnchannels(1)  # Assuming mono audio
            self.wf.setsampwidth(2)  # Assuming 16-bit audio
            self.wf.setframerate(44100)  # Assuming 44100Hz sample rate

    def transcribe(self, audio_bytes, platform, prompt="", language="en-US"):
        logger.info("Transcribing audio...")
        if platform == "web":
            audio = self._convert_webm_to_wav(audio_bytes, self.use == "local")
        else:
            audio = self._convert_bytes_to_wav(audio_bytes, self.use == "local")
        if self.use == "local":
            return self._transcribe(audio, prompt)
        elif self.use == "api":
            return self._transcribe_api(audio, prompt)

    def _transcribe(self, audio, prompt="", language="en-US"):
        language = WHISPER_LANGUAGE_CODE_MAPPING.get(language, config.language)
        segs, _ = self.model.transcribe(
            audio, language=language, vad_filter=True, initial_prompt=prompt
        )
        text = " ".join([seg.text for seg in segs])
        return text

    def _transcribe_api(self, audio, prompt=""):
        text = self.recognizer.recognize_whisper_api(
            audio,
            api_key=config.api_key,
        )
        return text

    def _convert_webm_to_wav(self, webm_data, local=True):
        webm_audio = AudioSegment.from_file(io.BytesIO(webm_data), format="webm")
        wav_data = io.BytesIO()
        webm_audio.export(wav_data, format="wav")
        if local:
            return wav_data
        with sr.AudioFile(wav_data) as source:
            audio = self.recognizer.record(source)
        return audio

    def _convert_bytes_to_wav(self, audio_bytes, local=True):
        if local:
            audio = io.BytesIO(sr.AudioData(audio_bytes, 44100, 2).get_wav_data())
            return audio
        return sr.AudioData(audio_bytes, 44100, 2)



from typing import List

from dotagent.schema import BaseRetriever, Document
from dotagent.knowledgebase.utilities.wikipedia import WikipediaAPIWrapper


class WikipediaRetriever(BaseRetriever, WikipediaAPIWrapper):
    """
    It is effectively a wrapper for WikipediaAPIWrapper.
    It wraps load() to get_relevant_documents().
    It uses all WikipediaAPIWrapper arguments without any change.
    """

    def get_relevant_documents(self, query: str) -> List[Document]:
        return self.load(query=query)

    async def aget_relevant_documents(self, query: str) -> List[Document]:
        raise NotImplementedError("wikipedia retriever does not support async")


"""Read PDF files."""

from pathlib import Path
from typing import Dict, List, Optional

from dotagent.knowledgebase.document_loaders.basereader import BaseReader
from dotagent.schema import DocumentNode


class PDFReader(BaseReader):
    """PDF reader."""

    def load_data(
        self, file: Path, extra_info: Optional[Dict] = None
    ) -> List[DocumentNode]:
        """Parse file."""
        import pypdf

        with open(file, "rb") as fp:
            # Create a PDF object
            pdf = pypdf.PdfReader(fp)

            # Get the number of pages in the PDF DocumentNode
            num_pages = len(pdf.pages)

            # Iterate over every page
            docs = []
            for page in range(num_pages):
                # Extract the text from the page
                page_text = pdf.pages[page].extract_text()
                page_label = pdf.page_labels[page]
                metadata = {"page_label": page_label, "file_name": file.name}

                if extra_info is not None:
                    metadata.update(extra_info)

                docs.append(DocumentNode(text=page_text, extra_info=metadata))
            return docs



from google.cloud import speech
import types

from dotagent.audio.speech_to_text.base import SpeechToText
from dotagent.logger import get_logger
from dotagent.utils import Singleton

logger = get_logger(__name__)
config = types.SimpleNamespace(**{
    'web': {
        'encoding': speech.RecognitionConfig.AudioEncoding.WEBM_OPUS,
        'sample_rate_hertz': 48000,
        'language_code': 'en-US',
        'max_alternatives': 1,
    },
    'terminal': {
        'encoding': speech.RecognitionConfig.AudioEncoding.LINEAR16,
        'sample_rate_hertz': 44100,
        'language_code': 'en-US',
        'max_alternatives': 1,
    },
})


class Google(Singleton, SpeechToText):
    def __init__(self):
        super().__init__()
        logger.info("Setting up [Google Speech to Text]...")
        self.client = speech.SpeechClient()

    def transcribe(self, audio_bytes, platform, prompt='', language='en-US') -> str:
        batch_config = speech.RecognitionConfig({
            'speech_contexts': [speech.SpeechContext(phrases=prompt.split(','))],
            **config.__dict__[platform]})
        batch_config.language_code = language
        if language != 'en-US':
            batch_config.alternative_language_codes = ['en-US']
        response = self.client.recognize(
            config=batch_config,
            audio=speech.RecognitionAudio(content=audio_bytes)
        )
        if not response.results:
            return ''
        result = response.results[0]
        if not result.alternatives:
            return ''
        return result.alternatives[0].transcript



from dotagent.schema import Document
from dotagent.tools.toolkits.slack_toolkit.slack.utils import SlackReader
from dotagent.tools.basetool import BaseTool
from typing import Optional, List, Type
from ssl import SSLContext
from pydantic import BaseModel, Field
from datetime import datetime
import logging

logger = logging.getLogger(__name__)

class SlackBase(BaseTool):

    def __init__(
            self,
            slack_token: Optional[str] = None,
            ssl: Optional[SSLContext] = None,
            earliest_date: Optional[datetime] = None,
            latest_date: Optional[datetime] = None,
        ) -> None:
            """Initialize with parameters."""
            self.reader = SlackReader(
                slack_token=slack_token,
                ssl=ssl,
                earliest_date=earliest_date,
                latest_date=latest_date,
            )

class LoadDataArgsSchema(BaseModel):
    channel_ids: List[str] = Field(
        ...,
        description=" Information about the parameter. ",
    )
    reverse_chronological: bool = Field(
        ...,
        description=" Information about the parameter.",
    )

class LoadData(SlackBase):
    name: str = "load_data"
    description: str = "Load data from the input directory."
    args_schema : Type[LoadDataArgsSchema] = LoadDataArgsSchema

    def load_data(
            self,
            channel_ids: List[str],
            reverse_chronological: bool = True,
        ) -> List[Document]:
            """Load data from the input directory."""
            return self.reader.load_data(
                channel_ids=channel_ids,
                reverse_chronological=reverse_chronological,
            )
    def _run(
            self,
            channel_ids: List[str],
            reverse_chronological: bool = True,
        )-> str:
         """Run the tool."""
         try:
            return self.load_data(self, channel_ids=channel_ids, reverse_chronological=reverse_chronological)
         except Exception as e:
              raise Exception(f"An error occurred: {e}")
        

class FetchChannel(SlackBase):
    name: str = "fetch_channel"
    description: str = "Fetch a list of relevant channels"
    def fetch_channels(
            self,
        ) -> List[str]:
            """Fetch a list of relevant channels."""
            slack_client = self.reader.client
            try:
                msg_result = slack_client.conversations_list()
                logger.info(msg_result)
            except Exception as e:
                logger.error(e)
                raise e

            return msg_result["channels"]
    def _run(
            self,
        )-> str:
         """Run the tool."""
         return self.fetch_channels()

class SendMessageArgsSchema(BaseModel):
    channel_id: str = Field(
        ...,
        description=" Information about the parameter. ",
    )
    message: str = Field(
        ...,
        description=" Information about the parameter. ",
    )
class SendMessage(SlackBase):
    name: str = "send_message"
    description: str = "Send a message to a channel given the channel ID."
    args_schema : Type[SendMessageArgsSchema] = SendMessageArgsSchema
    def send_message(
            self,
            channel_id: str,
            message: str,
        ) -> None:
            """Send a message to a channel given the channel ID."""
            slack_client = self.reader.client
            try:
                msg_result = slack_client.chat_postMessage(
                    channel=channel_id,
                    text=message,
                )
                logger.info(msg_result)
            except Exception as e:
                logger.error(e)
                raise e
    
    def _run(
            self,
            channel_id: str,
            message: str,
        )-> str:
         """Run the tool."""
         try:
            self.send_message(channel_id=channel_id, message=message)
            return "Message Sent"
         except Exception as e:
            raise Exception(f"An error occurred: {e}")
        


import asyncio
import os
import types
import httpx
import base64
from google.oauth2 import service_account
import google.auth.transport.requests

from dotagent.logger import get_logger
from dotagent.utils import Singleton
from dotagent.audio.text_to_speech.base import TextToSpeech

logger = get_logger(__name__)
DEBUG = False

config = types.SimpleNamespace(**{
    'url': 'https://texttospeech.googleapis.com/v1/text:synthesize',
    'headers': {
        'Content-Type': 'application/json',
    },
    'data': {
        'voice': {
            'languageCode': 'en-US',
            'name': 'en-US-Studio-M',
            'ssmlGender': 'NEUTRAL'
        },
        'audioConfig': {
            'audioEncoding': 'MP3'
        }
    },
    'service_account_file': os.getenv('GOOGLE_APPLICATION_CREDENTIALS', 'default/path.json'),
})


class GoogleCloudTTS(Singleton, TextToSpeech):
    def __init__(self):
        super().__init__()
        logger.info("Initializing [Google Cloud Text To Speech] voices...")

        # Load the service account key
        credentials = service_account.Credentials.from_service_account_file(
            config.service_account_file,
            scopes=["https://www.googleapis.com/auth/cloud-platform"]
        )

        # Request an access token
        auth_req = google.auth.transport.requests.Request()
        credentials.refresh(auth_req)

        # Now credentials.valid is True and credentials.token contains the access token
        self.access_token = credentials.token

        # Set the Authorization header with the access token
        config.headers['Authorization'] = f'Bearer {self.access_token}'

    async def stream(self, text, websocket, tts_event: asyncio.Event, voice_id="en-US-Standard-C",
                     first_sentence=False, language='en-US') -> None:
        if DEBUG:
            return
        if voice_id == "":
            logger.info("voice_id is not found in .env file, using Google default voice")
            voice_id = "en-US-Standard-C"
        headers = config.headers
        # For customized voices
        
        # if language != 'en-US':
        #     config.data["voice"]["languageCode"] = language
        #     config.data["voice"]["name"] = voice_id
        data = {
            "input": {
                "text": text
            },
            **config.data,
        }
        url = config.url
        async with httpx.AsyncClient() as client:
            response = await client.post(url, json=data, headers=headers)
            # Google Cloud TTS API does not support streaming, we send the whole content at once
            if response.status_code != 200:
                logger.error(f"Google Cloud TTS returns response {response.status_code}")
            else:
                audio_content = response.content
                # Decode the base64-encoded audio content
                audio_content = base64.b64decode(audio_content)
                await websocket.send_bytes(audio_content)



from importlib import metadata
from typing import Optional

# try:
#     __version__ = metadata.version(__package__)
# except metadata.PackageNotFoundError:
#     # Case where package metadata is not available.
#     __version__ = ""
# del metadata  # optional, avoids polluting the results of dir(__package__)

verbose: bool = False
debug: bool = False


from dotagent.tools.basetool import Tool
from dotagent.tools.toolkits.requests_toolkit.request import RequestsToolkit
from dotagent.tools.toolkits.requests_toolkit.requests.base import RequestsGetTool, RequestsPostTool, RequestsPatchTool, RequestsPutTool, RequestsDeleteTool
from dotagent.tools.toolkits.requests_toolkit.requests.utils import TextRequestsWrapper
from dotagent import compiler
import json

# requests_get_tool = RequestsGetTool(requests_wrapper=TextRequestsWrapper())
# requests_put_tool = RequestsPutTool(requests_wrapper=TextRequestsWrapper())
# requests_post_tool = RequestsPostTool(requests_wrapper=TextRequestsWrapper())
# requests_delete_tool = RequestsDeleteTool(requests_wrapper=TextRequestsWrapper())
# requests_patch_tool = RequestsPatchTool(requests_wrapper=TextRequestsWrapper())


request_tool = RequestsToolkit(requests_wrapper=TextRequestsWrapper())
request_tools = request_tool.get_tools()
# print(request_tools)

# requests_get_tool = Tool(
#     name = "Request_get",
#     func= requests_get_tool.run,
#     description="Useful when you have to get content from a URL"
# )

# requests_put_tool = Tool(
#     name = "Request_put",
#     func= requests_put_tool.run,
#     description = """Use this when you want to PUT to a website.
#     Input should be a json string with two keys: "url" and "data".
#     The value of "url" should be a string, and the value of "data" should be a dictionary of 
#     key-value pairs you want to PUT to the url.
#     Be careful to always use double quotes for strings in the json string.
#     The output will be the text response of the PUT request.
#     """
# )

# requests_delete_tool = Tool(
#     name = "Request_delete",
#     func= requests_delete_tool.run,
#     description="Useful when you have to make delete request to a URL"
# )

# requests_patch_tool = Tool(
#     name = "Request_patch",
#     func= requests_patch_tool.run,
#     description="""Use this when you want to PATCH to a website.
#     Input should be a json string with two keys: "url" and "data".
#     The value of "url" should be a string, and the value of "data" should be a dictionary of 
#     key-value pairs you want to PATCH to the url.
#     Be careful to always use double quotes for strings in the json string
#     The output will be the text response of the PATCH request.
#     """
# )

# requests_post_tool = Tool(
#     name = "Request_post",
#     func= requests_post_tool.run,
#     description="""Use this when you want to POST to a website.
#     Input should be a json string with two keys: "url" and "data".
#     The value of "url" should be a string, and the value of "data" should be a dictionary of 
#     key-value pairs you want to POST to the url.
#     Be careful to always use double quotes for strings in the json string
#     The output will be the text response of the POST request.
#     """
# )

# tools = [requests_get_tool,requests_put_tool,requests_delete_tool,requests_patch_tool, requests_post_tool]

# we use GPT-4 here, but you could use gpt-3.5-turbo as well
llm = compiler.llms.OpenAI(model="gpt-3.5-turbo-16k")

def tool_use(query, tools=request_tools):
    query = json.loads(query)
    return tools[int(query["index"])].run(query["query"])

experts = compiler(template='''
{{#system~}}
You are a helpful Web assistant. You are given a set of tools to use
{{~#each tools}}
{{this}}
{{/each}}
{{~/system}}

{{#user~}}
I want a response to the following question:
{{query}}
Think do you need to use the given tool to answer the question. Provide the answer in either <<Yes>> or <<No>>.
{{~/user}}

{{#assistant~}}
{{gen 'tools_use' temperature=0 max_tokens=300}}
{{~/assistant}}
                        
{{#user~}}
If the answer is Yes then call the tool using the following format '{"index":[index of the tool to be used in the tools list], "query":[query to be passed]'
If the answer is No, answer to the {{query}} itself.
{{~/user}}

{{#assistant~}}
{{gen 'action' temperature=0 max_tokens=500}}
{{#if (tools_use)=="Yes"}}
{{(tool_func action)}}
{{/if}}
{{~/assistant}}    
                        
{{#user~}}
Summarise the answer in one sentence
{{~/user}}
                        
{{#assistant~}}
{{gen 'final_answer' temperature=0 max_tokens=500}}
{{~/assistant}}
''', 
llm=llm, tools = request_tools, tool_func = tool_use, stream = False)

# get request (Gets the response provided by the given URL)
out = experts(query='Get the content from the following URL : "https://yaml.org"')
print(out)

#All the below tools result in "405 bad gateway, Not Allowed" response
#delete request
# out = experts(query='Make a DELETE request to the following URL : "https://yaml.org"')
# print(out)

# put request
# out = experts(query="""Put to the following URL with the following data {
#   "url": "http://yaml.org",
#   "data": {
#     "key1": "value1",
#     "key2": "value2",
#     "key3": "value3"
#   }
# }""")
# print(out)

# Post request
# out = experts(query="""Post to the following URL with the following data {
#   "url": "http://yaml.org",
#   "data": {
#     "key1": "value1",
#     "key2": "value2",
#     "key3": "value3"
#   }
# }""")
# print(out)

# Patch request
# out = experts(query="""Patch to the following URL with the following data {
#   "url": "http://yaml.org",
#   "data": {
#     "key1": "value1",
#     "key2": "value2",
#     "key3": "value3"
#   }
# }""")
# print(out)



import click
from pathlib import Path
import logging
from dotagent.gui.cli import run_gui  # Ensure this import is correct

# Configure logging
logging.basicConfig(level=logging.DEBUG)

@click.group()
def cli():
    pass

@click.command()
@click.option('--gui', is_flag=True, help='Run the GUI code.')
@click.argument('filename', type=Path)
def run(gui, filename):
    try:
        logging.info(f"Running command with GUI flag: {gui} and filename: {filename}")
        
        if gui:
            logging.info(f"Running GUI with file {filename}")
            run_gui(filename)  # Ensure this function is implemented
            logging.info("GUI run completed.")
            print("GUI run completed.")  # Feedback to user
        else:
            logging.info(f"Running {filename} without GUI")
            print(f"Running {filename} without GUI")
            # TODO: Add the code that runs the script without the GUI
        
    except Exception as e:
        logging.error(f"An error occurred: {e}")
        print(f"An error occurred: {e}")  # Feedback to user

cli.add_command(run)

if __name__ == '__main__':
    logging.info("Starting CLI.")
    cli()
    logging.info("CLI execution completed.")


from enum import Enum
from typing import List, Dict, Any, Union, Optional
import os
from dotagent import compiler
from dotagent.tools.basetool import BaseTool
from dotagent.agent.base_agent import BaseAgent
import logging
import yaml
import argparse

class AgentState(Enum):
    """Enum to represent different states of the Agent"""

    IDLE = 0
    BUSY = 1
    USED_AS_TOOL = 2
    ERROR = 3

class TextCompletionAgent(BaseAgent):
    def __init__(self,
                knowledgebase: Optional[Any] = None,
                tools: Optional[Any] = None,
                llm: Optional[Any] = None,
                prompt_template: str = "{{input}} {{gen 'output'}}",
                input_variables: Dict = {},
                agent_id: str = "default",
                memory: Any = None,
                caching: bool = False,
                output_key: str = None,
                return_complete: bool = False
                ):
        
        self.agent_id = agent_id
        self.knowledgebase = knowledgebase
        self.tools = tools
        self.prompt_template = prompt_template
        self.input_variables = input_variables
        self.memory = memory
        self.state = AgentState.IDLE
        self.caching = caching
        self.return_complete = return_complete
        self.llm = llm if llm is not None else self.llm_instance()

        self.parser = argparse.ArgumentParser(description="CLI for interacting with the Agent instance.")

        self.output_key = output_key
        self.compiler = compiler(llm = self.llm, template = self.prompt_template, caching=caching)

        if self.return_complete and self.output_key is not None:
            logging.warning("return_complete mode is enabled. Output key will be ignored.")

        if self.knowledgebase is not None and not self.input_variables.get('knowledge_variable'):
            raise ValueError ("knowledge_variable should be present in input_variables while using knowledge")
        

    @property
    def agent_type(self):
        return "completion"

    @property
    def get_knowledge_variable(self):
        """Get knowledge variable name from input variables"""
        return self.input_variables.get('knowledge_variable')
    
    @property
    def default_llm_model(self):
        return "text-davinci-003"
        
    def add_tool(self, tool: BaseTool) -> None:
        """Add a tool to the agent's tool list."""
        self.tools.append(tool)

    def remove_tool(self, tool: BaseTool) -> None:
        """Remove a tool from the agent's tool list."""
        if tool in self.tools:
            self.tools.remove(tool)

    def llm_instance(self) -> compiler.llms.OpenAI:
        """Create an instance of the language model."""
        return compiler.llms.OpenAI(model=self.default_llm_model)
    
    def get_output_key(self, prompt):
        vars = prompt.variables()
        vars = list(vars.keys())
        return vars[-1]

    def get_knowledge(self, query) -> List[str]:
        docs = self.knowledgebase.retrieve_data(query)
        final_doc = ""
        for doc in docs:
            final_doc += doc
        return final_doc

    



import re
import spacy

class PIIScrubber:
    """
    Class for scrubbing personally identifiable information (PII) from text.
    """
    PATTERNS = {
        'email': r'\b[A-Za-z0-9._%+-]+@[A-Za-z0-9.-]+\.[A-Z|a-z]{2,7}\b',
        'ssn': r'\d{3}-?\d{2}-?\d{4}',
        'credit_card': r'\d{4}-?\d{4}-?\d{4}-?\d{4}',
        'phone_number': r'\b(\+\d{1,2}\s)?\(?\d{3}\)?[\s.-]?\d{3}[\s.-]?\d{4}\b',
        'ip': r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}',
        'date_of_birth': r'\b(0[1-9]|1[0-2])[- /.](0[1-9]|[12][0-9]|3[01])[- /.](19|20)\d\d\b',
        'vehicle_identification_number': r'\b([A-HJ-NPR-Z0-9]{3})([A-HJ-NPR-Z0-9]{5})(\d{2})([A-HJ-NPR-Z0-9]{8})\b',
    }

    def __init__(self, verbose=False):
        """
        Initialize the PIIScrubber with given verbosity.

        Args:
            verbose (bool): If True, print out internal states. Default is False.
        """
        self.verbose = verbose
        self.nlp = spacy.load('en_core_web_sm')

    def scrub(self, text):
        """
        Detect and log PII in the given text based on regex patterns and SpaCy NER.
    
        Args:
            text (str): The input text to scrub.

        Returns:
            dict: A dictionary with the start and end indices of the detected PII as keys and the corresponding replacements as values.
        """
        replacements = {}
        for label, pattern in self.PATTERNS.items():
            for match in re.finditer(pattern, text):
                replacements[match.span()] = f"[REDACTED {label.upper()}]"
                if self.verbose:
                    print(f"Potential {label} detected: {match.group()}")

        doc = self.nlp(text)
        for ent in doc.ents:
            if ent.label_ in ['PERSON', 'ORG']:
                replacements[ent.start_char, ent.end_char] = "[REDACTED NAME]"
                if self.verbose:
                    print(f"Potential {ent.label_} entity detected: {ent.text}")

        return replacements

    @staticmethod
    def anonymize_text(text, replacements):
        """
        Anonymize PII in the given text based on the provided replacements.
    
        Args:
            text (str): The input text to anonymize.
            replacements (dict): A dictionary with the start and end indices of the PII to replace as keys and the corresponding replacements as values.

        Returns:
            str: The anonymized text.
        """
        replacements = sorted(replacements.items(), key=lambda x: x[0][0], reverse=True)
        for (start, end), replacement in replacements:
            text = text[:start] + replacement + text[end:]

        return text

    def run(self, text):
        """
        Detect, log, and anonymize PII in the given text.
    
        Args:
            text (str): The input text to scrub.

        Returns:
            str: The anonymized text.
        """
        if not isinstance(text, str):
            raise TypeError('Text must be a string')

        try:
            replacements = self.scrub(text)
            anonymized_text = self.anonymize_text(text, replacements)
            if text != anonymized_text and self.verbose:
                print(f'Original: {text}\nAnonymized: {anonymized_text}')

        except Exception as e:
            print(f"Error processing text: {e}")
            raise e

        return anonymized_text



from dotagent.tools.basetool import BaseTool
from dotagent.schema import Document
from pydantic import BaseModel, Field
import datetime
import os
from typing import Any, List, Optional, Union, Type


SCOPES = ["https://www.googleapis.com/auth/calendar"]


class GoogleCalendar(BaseTool):
    """Google Calendar tool spec.

    Currently a simple wrapper around the data loader.
    TODO: add more methods to the Google Calendar tool.

    """
def _get_credentials(self) -> Any:
    """Get valid user credentials from storage.

    The file token.json stores the user's access and refresh tokens, and is
    created automatically when the authorization flow completes for the first
    time.

    Returns:
        Credentials, the obtained credential.
    """
    from google.auth.transport.requests import Request
    from google.oauth2.credentials import Credentials
    from google_auth_oauthlib.flow import InstalledAppFlow

    creds = None
    if os.path.exists("token.json"):
        creds = Credentials.from_authorized_user_file("token.json", SCOPES)
    # If there are no (valid) credentials available, let the user log in.
    if not creds or not creds.valid:
        if creds and creds.expired and creds.refresh_token:
            creds.refresh(Request())
        else:
            flow = InstalledAppFlow.from_client_secrets_file(
                "credentials.json", SCOPES
            )
            creds = flow.run_local_server(port=8080)
        # Save the credentials for the next run
        with open("token.json", "w") as token:
            token.write(creds.to_json())

    return creds

class LoadDataArgsSchema(BaseModel):
    number_of_results: Optional[int] = Field(
        ...,
        description=" Information about the parameter. ",
    )
    start_date: Optional[Union[str, datetime.date]] = Field(
        ...,
        description=" Information about the parameter. ",
    )

class LoadData(GoogleCalendar):
    name : str = "Load Data"
    description : str = "Load data from user's calendar."
    args_schema : Type[LoadDataArgsSchema] = LoadDataArgsSchema
     
    def load_data(
            self,
            number_of_results: Optional[int] = 100,
            start_date: Optional[Union[str, datetime.date]] = None,
        ) -> List[Document]:

            """Load data from user's calendar.

            Args:
                number_of_results (Optional[int]): the number of events to return. Defaults to 100.
                start_date (Optional[Union[str, datetime.date]]): the start date to return events from in date isoformat. Defaults to today.
            """

            from googleapiclient.discovery import build

            credentials = _get_credentials()
            service = build("calendar", "v3", credentials=credentials)

            if start_date is None:
                start_date = datetime.date.today()
            elif isinstance(start_date, str):
                start_date = datetime.date.fromisoformat(start_date)

            start_datetime = datetime.datetime.combine(start_date, datetime.time.min)
            start_datetime_utc = start_datetime.strftime("%Y-%m-%dT%H:%M:%S.%fZ")

            events_result = (
                service.events()
                .list(
                    calendarId="primary",
                    timeMin=start_datetime_utc,
                    maxResults=number_of_results,
                    singleEvents=True,
                    orderBy="startTime",
                )
                .execute()
            )

            events = events_result.get("items", [])

            if not events:
                return []

            results = []
            for event in events:
                if "dateTime" in event["start"]:
                    start_time = event["start"]["dateTime"]
                else:
                    start_time = event["start"]["date"]

                if "dateTime" in event["end"]:
                    end_time = event["end"]["dateTime"]
                else:
                    end_time = event["end"]["date"]

                event_string = f"Status: {event['status']}, "
                event_string += f"Summary: {event['summary']}, "
                event_string += f"Start time: {start_time}, "
                event_string += f"End time: {end_time}, "

                organizer = event.get("organizer", {})
                display_name = organizer.get("displayName", "N/A")
                email = organizer.get("email", "N/A")
                if display_name != "N/A":
                    event_string += f"Organizer: {display_name} ({email})"
                else:
                    event_string += f"Organizer: {email}"

                results.append(Document(text=event_string))

            return results
    def run(
         self,
            number_of_results: Optional[int] = 100,
            start_date: Optional[Union[str, datetime.date]] = None,
    )->str:
        try:
            return self.load_data(self, number_of_results=number_of_results, start_date=start_date)
        except Exception as e:
            raise Exception(f"An error occurred: {e}")

class CreateEventArgsSchema(BaseModel):
    title: Optional[str] = Field(
        ...,
        description=" Information about the parameter. ",
    )
    description: Optional[str] = Field(
        ...,
        description=" Information about the parameter. ",
    )
    location: Optional[str] = Field(
        ...,
        description=" Information about the parameter. ",
    )
    start_datetime: Optional[Union[str, datetime.datetime]] = Field(
        ...,
        description=" Information about the parameter. ",
    )
    end_datetime: Optional[Union[str, datetime.datetime]] = Field(
        ...,
        description=" Information about the parameter. ",
    )
    attendees: Optional[List[str]] = Field(
        ...,
        description=" Information about the parameter. ",
    )

class CreateEvent(GoogleCalendar):
    name : str = "Create Event"
    description : str = "Create an event on the users calendar."
    args_schema : Type[CreateEventArgsSchema] = CreateEventArgsSchema

    def create_event(
        self,
        title: Optional[str] = None,
        description: Optional[str] = None,
        location: Optional[str] = None,
        start_datetime: Optional[Union[str, datetime.datetime]] = None,
        end_datetime: Optional[Union[str, datetime.datetime]] = None,
        attendees: Optional[List[str]] = None,
    ) -> str:
        """
            Create an event on the users calendar

        Args:
            title (Optional[str]): The title for the event
            description (Optional[str]): The description for the event
            location (Optional[str]): The location for the event
            start_datetime Optional[Union[str, datetime.datetime]]: The start datetime for the event
            end_datetime Optional[Union[str, datetime.datetime]]: The end datetime for the event
            attendees Optional[List[str]]: A list of email address to invite to the event
        """

        from googleapiclient.discovery import build

        credentials = _get_credentials()
        service = build("calendar", "v3", credentials=credentials)

        attendees_list = []
        for attendee in attendees:
            attendees_list.append({"email": attendee})
        start_time = (
            datetime.datetime.strptime(start_datetime, "%Y-%m-%dT%H:%M:%S")
            .astimezone()
            .strftime("%Y-%m-%dT%H:%M:%S.%f%z")
        )
        end_time = (
            datetime.datetime.strptime(end_datetime, "%Y-%m-%dT%H:%M:%S")
            .astimezone()
            .strftime("%Y-%m-%dT%H:%M:%S.%f%z")
        )

        event = {
            "summary": title,
            "location": location,
            "description": description,
            "start": {
                "dateTime": start_time,
            },
            "end": {
                "dateTime": end_time,
            },
            "attendees": attendees_list,
        }
        event = service.events().insert(calendarId="primary", body=event).execute()
        return "Your calendar event has been created successfully! You can move on to the next step."

    def run(
         self,
        title: Optional[str] = None,
        description: Optional[str] = None,
        location: Optional[str] = None,
        start_datetime: Optional[Union[str, datetime.datetime]] = None,
        end_datetime: Optional[Union[str, datetime.datetime]] = None,
        attendees: Optional[List[str]] = None,   
    )-> str:
        try:
            return self.create_event(self, title=title, description=description, location=location, start_datetime=start_datetime,end_datetime=end_datetime,attendees=attendees)
        except Exception as e:
              raise Exception(f"An error occurred: {e}")

class GetDate(GoogleCalendar):
    name : str = "Get date"
    description : str = "A function to return todays date. Call this before any other functions if you are unaware of the date."

    def get_date(self):
        """
        A function to return todays date. Call this before any other functions if you are unaware of the date
        """
        return datetime.date.today()
    
    def run(self):
        return(self.get_date())